Package: origami
Title: Generalized Framework for Cross-Validation
Version: 1.0.8
Authors@R: c(
    person("Jeremy", "Coyle", email = "jeremyrcoyle@gmail.com",
           role = c("aut", "cph"),
           comment = c(ORCID = "0000-0002-9874-6649")),
    person("Nima", "Hejazi", email = "nh@nimahejazi.org",
           role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-7127-2789")),
    person("Ivana", "Malenica", email = "imalenica@berkeley.edu",
           role = "aut",
           comment = c(ORCID = "0000-0002-7404-8088")),
    person("Rachael", "Phillips", email = "rachaelvphillips@berkeley.edu",
           role = "aut",
           comment = c(ORCID = "0000-0002-8474-591X"))
  )
Maintainer: Nima Hejazi <nh@nimahejazi.org>
Description: A general framework for the application of cross-validation schemes
    to particular functions. By allowing arbitrary lists of results, origami
    accommodates a range of cross-validation applications. This implementation
    was first described by Coyle and Hejazi (2018) <doi:10.21105/joss.00512>.
Depends: R (>= 3.0.0),
License: GPL-3
URL: https://tlverse.org/origami/
BugReports: https://github.com/tlverse/origami/issues
Encoding: UTF-8
Imports: abind, methods, data.table, assertthat, future, future.apply,
        listenv
Suggests: testthat, class, rmarkdown, knitr, stringr, glmnet, forecast,
        randomForest
VignetteBuilder: knitr
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-05-12 20:48:26 UTC; nih914
Author: Jeremy Coyle [aut, cph] (ORCID:
    <https://orcid.org/0000-0002-9874-6649>),
  Nima Hejazi [aut, cre] (ORCID: <https://orcid.org/0000-0002-7127-2789>),
  Ivana Malenica [aut] (ORCID: <https://orcid.org/0000-0002-7404-8088>),
  Rachael Phillips [aut] (ORCID: <https://orcid.org/0000-0002-8474-591X>)
Repository: CRAN
Date/Publication: 2026-05-18 05:50:11 UTC
Built: R 4.5.2; ; 2026-05-18 08:47:17 UTC; unix

# Generated by roxygen2: do not edit by hand

export(combine_results)
export(combiner_array)
export(combiner_c)
export(combiner_factor)
export(combiner_rbind)
export(cross_validate)
export(fold_from_foldvec)
export(fold_index)
export(folds2foldvec)
export(folds_bootstrap)
export(folds_loo)
export(folds_montecarlo)
export(folds_resubstitution)
export(folds_rolling_origin)
export(folds_rolling_origin_pooled)
export(folds_rolling_window)
export(folds_rolling_window_pooled)
export(folds_vfold)
export(folds_vfold_rolling_origin_pooled)
export(folds_vfold_rolling_window_pooled)
export(id_folds_to_folds)
export(make_fold)
export(make_folds)
export(make_repeated_folds)
export(training)
export(validation)
export(wrap_in_try)
import(methods)
importFrom(abind,abind)
importFrom(assertthat,assert_that)
importFrom(data.table,rbindlist)
importFrom(future,future)
importFrom(future,multicore)
importFrom(future,multisession)
importFrom(future,plan)
importFrom(future,sequential)
importFrom(future.apply,future_lapply)
importFrom(listenv,listenv)

Origami
=====
[![Gem Version](https://badge.fury.io/rb/origami.svg)](http://rubygems.org/gems/origami)

Overview
--------

Origami is a framework written in pure Ruby to manipulate PDF files.

It offers the possibility to parse the PDF contents, modify and save the PDF
structure, as well as creating new documents.

Origami supports some advanced features of the PDF specification:

* Compression filters with predictor functions
* Encryption using RC4 or AES, including the undocumented Revision 6 derivation algorithm
* Digital signatures and Usage Rights
* File attachments
* AcroForm and XFA forms
* Object streams

Origami is able to parse PDF, FDF and PPKLite (Adobe certificate store) files.

Requirements
------------

As of version 2, the minimal version required to run Origami is Ruby 2.1.

Some optional features require additional gems:

* [Gtk2][ruby-gtk2] for running the GUI interface
* [therubyracer][the-ruby-racer] for JavaScript emulation of PDF scripts

[ruby-gtk2]: https://rubygems.org/gems/gtk2
[the-ruby-racer]: https://rubygems.org/gems/therubyracer

Quick start
-----------

First install Origami using the latest gem available:

    $ gem install origami

Then import Origami with:

```ruby
require 'origami'
include Origami
```

To process a PDF document, you can use the ``PDF.read`` method:

```ruby
pdf = PDF.read "something.pdf"

puts "This document has #{pdf.pages.size} page(s)"
```

The default behavior is to parse the entire contents of the document at once. This can be changed by passing the ``lazy`` flag to parse objects on demand.

```ruby
pdf = PDF.read "something.pdf", lazy: true

pdf.each_page do |page|
    page.each_font do |name, font|
        # ... only parse the necessary bits
    end
end
```

You can also create documents directly by instanciating a new PDF object:

```ruby
pdf = PDF.new

pdf.append_page
pdf.pages.first.write "Hello", size: 30

pdf.save("example.pdf")

# Another way of doing it
PDF.write("example.pdf") do |pdf|
    pdf.append_page do |page|
        page.write "Hello", size: 30
    end
end
```

Take a look at the [examples](examples) and [bin](bin) directories for some examples of advanced usage.

Tools
-----

Origami comes with a set of tools to manipulate PDF documents from the command line.

* [pdfcop](bin/pdfcop): Runs some heuristic checks to detect dangerous contents.
* [pdfdecompress](bin/pdfdecompress): Strips compression filters out of a document.
* [pdfdecrypt](bin/pdfdecrypt): Removes encrypted contents from a document.
* [pdfencrypt](bin/pdfencrypt): Encrypts a PDF document.
* [pdfexplode](bin/pdfexplode): Explodes a document into several documents, each of them having one deleted resource. Useful for reduction of crash cases after a fuzzing session.
* [pdfextract](bin/pdfextract): Extracts binary resources of a document (images, scripts, fonts, etc.).
* [pdfmetadata](bin/pdfmetadata): Displays the metadata contained in a document.
* [pdf2ruby](bin/pdf2ruby): Converts a PDF into an Origami script rebuilding an equivalent document (experimental).
* [pdfsh](bin/pdfsh): An IRB shell running inside the Origami namespace.
* [pdfwalker](bin/pdfwalker): A graphical interface to dig into the contents of a PDF document.


License
-------

Origami is distributed under the [LGPL](COPYING.LESSER) license, except for the graphical interface which is distributed under the [GPL](bin/gui/COPYING) license.

Copyright © 2016 Guillaume Delugré.

# Origami: Analytical GEMM Solution Selection

The name "origami" still evokes the elegance of transforming a flat (2-D) sheet into intricate higher dimensional structures. In this context, however, Origami has evolved into a toolset for **GEMM solution selection and optimization**. Inspired by the art of paper folding, the library now enables users to explore a range of tiling and mapping configurations and to make informed decisions on data and computation mapping for high-performance GEMM operations.

Origami provides a rigorous methodology to analyze and select optimal GEMM parameters. It does so by evaluating both **compute** and **memory latencies** across a wide range of tile sizes. The framework computes essential metrics such as:
- **Matrix Instruction (MI) Tiling Counts** and the associated compute latencies.
- **Memory Load Latencies**, considering both L2 and main memory bandwidth constraints.
- **Active Compute Unit Occupancy and Wave Counts**, ensuring realistic mapping to the underlying hardware.
- **Tie-breaking Strategies** using arithmetic intensity and more refined L2 hit rate estimations to resolve candidate configuration ties.

This approach allows programmers to achieve near-optimal performance without manually exploring every possibility. While default (implicit) parameters provide "out-of-the-box" performance, expert users can dive deeper into the analytical model to tune for their specific hardware and problem sizes.

---

## Documentation

- **Getting Started**
- [Programming Abstraction]()
- [Hierarchical Tiling and GEMM Latency Calculation]()
- [Usage Examples]()

---

## Quick Start Guide – Origami

**Origami** provides an end-to-end analytical solution to GEMM parameter selection. It estimates performance by sweeping over candidate tile sizes and selecting the optimal configuration based on latency and arithmetic intensity.

### Building Origami

Assuming you are in the repository root, run:

```bash
# configure
cmake -S . -B build/ -DCMAKE_PREFIX_PATH=/opt/rocm -DCMAKE_CXX_COMPILER=/opt/rocm/bin/amdclang++ -DCMAKE_INSTALL_PREFIX=/opt/rocm 
# build
cmake --build build/ --parallel
```

### Installing Origami

After configuring and building, run the following command to install:

```bash
# install
cmake --target install
```

### Testing Origami Python Bindings

To test the origami Python bindings:

```bash
# configure with python bindings and tests enabled 
cmake -S . -B build/ -DCMAKE_PREFIX_PATH=/opt/rocm -DCMAKE_CXX_COMPILER=/opt/rocm/bin/amdclang++ -DCMAKE_INSTALL_PREFIX=/opt/rocm -D ORIGAMI_ENABLE_PYTHON=ON -D ORIGAMI_BUILD_TESTING=ON

# build 
cmake --build build/ --parallel

# run tests
cd build/
ctest --output-on-failure
```

### Options
* `ORIGAMI_BUILD_SHARED_LIBS`: Enables building of shared libraries (default: `ON`)
* `ORIGAMI_ENABLE_PYTHON`: Enables generation of origami Python bindings (default: `OFF`)
* `ORIGAMI_BUILD_TESTING`: Build the Python binding tests (default: `OFF`)


### Building and Running Origami C++ Tests
Use CMake to configure the build system and compile Origami with testing enabled. From origami root `<rocm-libraries-root>/shared/origami`:

```bash
cmake -S . -B build/ \
  -DCMAKE_PREFIX_PATH=/opt/rocm \
  -DCMAKE_CXX_COMPILER=/opt/rocm/bin/amdclang++ \
  -DCMAKE_INSTALL_PREFIX=/opt/rocm \
  -D ORIGAMI_BUILD_TESTING=ON

cmake --build build/ --parallel
```

Once the build completes, navigate to the test directory and run the test suite:

```bash
cd build/tests/
./origami-tests
```

# Origami Python Bindings

This directory provides Python bindings for the C++ Origami project.
The bindings allow you to use Origami's functionality directly from Python.

## Installation

First, install `nanobind` using pip:

```bash
pip install nanobind
```

Make sure `ROCM_PATH` is set (e.g. `export ROCM_PATH=/opt/rocm`). Then, build the bindings using the following command:

```bash
python setup.py build_ext --inplace
export PYTHONPATH=$(pwd):$PYTHONPATH
```

## Testing

After building the bindings, you can test them using the provided test file:

```bash
python ./origami_test.py
```

## Usage

Once the bindings are built and tested, you can import and use them in your Python scripts:

```python
import origami

hardware = origami.getHardwareForDevice(args.device)

result = origami.select_best_macro_tile_size(
            args.m, args.n, args.k, args.transA, args.transB, hardware, tile_list, args.element_size, args.miDataType, 0.8, args.debug, args.print, args.wgm
        )
```

## Modifying `origami_module.cpp`

To expose more C++ functionality to Python, you can modify the `origami_module.cpp` file.
This involves adding new functions or classes to the module and ensuring they are properly wrapped for Python.
After making your changes, rebuild the bindings using the installation command above.

###############################################################################
# This example explains how to use the cross_validate function naively.
###############################################################################
data(mtcars)

# resubstitution MSE
r <- lm(mpg ~ ., data = mtcars)
mean(resid(r)^2)

# function to calculate cross-validated squared error
cv_lm <- function(fold, data, reg_form) {
  # get name and index of outcome variable from regression formula
  out_var <- as.character(unlist(stringr::str_split(reg_form, " "))[1])
  out_var_ind <- as.numeric(which(colnames(data) == out_var))

  # split up data into training and validation sets
  train_data <- training(data)
  valid_data <- validation(data)

  # fit linear model on training set and predict on validation set
  mod <- lm(as.formula(reg_form), data = train_data)
  preds <- predict(mod, newdata = valid_data)

  # capture results to be returned as output
  out <- list(
    coef = data.frame(t(coef(mod))),
    SE = ((preds - valid_data[, out_var_ind])^2)
  )
  return(out)
}

# replicate the resubstitution estimate
resub <- make_folds(mtcars, fold_fun = folds_resubstitution)[[1]]
resub_results <- cv_lm(fold = resub, data = mtcars, reg_form = "mpg ~ .")
mean(resub_results$SE)

# cross-validated estimate
folds <- make_folds(mtcars)
cv_results <- cross_validate(
  cv_fun = cv_lm, folds = folds, data = mtcars,
  reg_form = "mpg ~ ."
)
mean(cv_results$SE)

cvfun <- function(split, aux = NULL, time = time) {
  code <- sprintf("t:%s s:%s", time, split$v)
  if (is.null(aux)) {
    sprintf("%s", code)
  } else {
    sprintf("%s - %s", code, aux[split$v])
  }
}

test1 <- function(splits, time) {
  if (time > 0) {
    pastresults <- test1(splits, time = time - 1)
    aux <- pastresults$splitResults
  } else {
    aux <- NULL
  }
  # t2
  splitResults <- genCV(cvfun, splits,
    aux = aux, .doPar = T,
    time = time
  )[[1]]
  results <- list(full = sprintf("t:%s full", time), splitResults = splitResults)
  return(results)
}

test1(splits, time = 2)

test2 <- function(splits, time) {
  if (time > 0) {
    pastresults <- test2(splits, time = time - 1)
    aux <- rep(pastresults$full, length(splits))
  } else {
    aux <- NULL
  }
  # t2
  splitResults <- genCV(cvfun, splits,
    aux = aux, .doPar = T,
    time = time
  )[[1]]
  results <- list(full = sprintf("t:%s full", time), splitResults = splitResults)
  return(results)
}

test2(splits, time = 2)

testonsplit <- function(split, time) {
  newsplits <- gensplits(split$test_idx)
  results <- test3(newsplits, time)
  return(results)
}

test3 <- function(splits, time) {
  if (time > 0) {
    pastSplitResults <- genCV(testonsplit, splits,
      .doPar = T,
      time = time - 1
    )
    aux <- pastSplitResults$full
    aux <- paste(aux, "nested")
  } else {
    aux <- NULL
  }
  # t2
  splitResults <- genCV(cvfun, splits,
    aux = aux, .doPar = T,
    time = time
  )[[1]]
  results <- list(full = sprintf("t:%s full", time), splitResults = splitResults)
  return(results)
}

test3(splits, time = 2)

strataids <- rep(1:2, 20)
n <- length(strataids)
folds <- make_folds(strataids = strataids)
lapply(folds, function(fold) {
  table(strataids, seq_len(n) %in%
    training(fold = fold))
})

clusterids <- rep(1:20, 2)
n <- length(clusterids)
folds <- make_folds(clusterids = clusterids)
lapply(folds, function(fold) {
  table(clusterids, seq_len(n) %in%
    training(fold = fold))
})

folds <- make_folds(V = 2, strataids = strataids, clusterids = clusterids)

lapply(folds, function(fold) {
  table(strataids, clusterids, seq_len(n) %in%
    training(fold = fold))
})

## ----load_data----------------------------------------------------------------
data(mtcars)
head(mtcars)

## ----linear_mod---------------------------------------------------------------
lm_mod <- lm(mpg ~ ., data = mtcars)
summary(lm_mod)

## ----get_naive_error----------------------------------------------------------
err <- mean(resid(lm_mod)^2)

## ----define_fun_cv_lm---------------------------------------------------------
cv_lm <- function(fold, data, reg_form) {
  # get name and index of outcome variable from regression formula
  out_var <- as.character(unlist(str_split(reg_form, " "))[1])
  out_var_ind <- as.numeric(which(colnames(data) == out_var))

  # split up data into training and validation sets
  train_data <- training(data)
  valid_data <- validation(data)

  # fit linear model on training set and predict on validation set
  mod <- lm(as.formula(reg_form), data = train_data)
  preds <- predict(mod, newdata = valid_data)

  # capture results to be returned as output
  out <- list(
    coef = data.frame(t(coef(mod))),
    SE = ((preds - valid_data[, out_var_ind])^2)
  )
  return(out)
}

## ----load_pkgs----------------------------------------------------------------
library(origami)
library(stringr) # used in defining the cv_lm function above

## ----cv_lm_resub--------------------------------------------------------------
# resubstitution estimate
resub <- make_folds(mtcars, fold_fun = folds_resubstitution)[[1]]
resub_results <- cv_lm(fold = resub, data = mtcars, reg_form = "mpg ~ .")
mean(resub_results$SE)

## ----cv_lm_cross_valdate------------------------------------------------------
# cross-validated estimate
folds <- make_folds(mtcars)
cvlm_results <- cross_validate(
  cv_fun = cv_lm,
  folds = folds,
  data = mtcars,
  reg_form = "mpg ~ ."
)
mean(cvlm_results$SE)

## ----cv_fun_randomForest------------------------------------------------------
cv_rf <- function(fold, data, reg_form) {
  # get name and index of outcome variable from regression formula
  out_var <- as.character(unlist(str_split(reg_form, " "))[1])
  out_var_ind <- as.numeric(which(colnames(data) == out_var))

  # define training and validation sets based on input object of class "folds"
  train_data <- training(data)
  valid_data <- validation(data)

  # fit Random Forest regression on training set and predict on holdout set
  mod <- randomForest(formula = as.formula(reg_form), data = train_data)
  preds <- predict(mod, newdata = valid_data)

  # define output object to be returned as list (for flexibility)
  out <- list(
    coef = data.frame(mod$coefs),
    SE = ((preds - valid_data[, out_var_ind])^2)
  )
  return(out)
}

## -----------------------------------------------------------------------------
library(randomForest)
folds <- make_folds(mtcars)
cvrf_results <- cross_validate(
  cv_fun = cv_rf,
  folds = folds,
  data = mtcars,
  reg_form = "mpg ~ ."
)
mean(cvrf_results$SE)

## -----------------------------------------------------------------------------
data(AirPassengers)
print(AirPassengers)

## -----------------------------------------------------------------------------
library(forecast)
folds = make_folds(
  AirPassengers,
  fold_fun = folds_rolling_origin,
  first_window = 36,
  validation_size = 24
)
fold = folds[[1]]

# function to calculate cross-validated squared error
cv_forecasts <- function(fold, data) {
  train_data <- training(data)
  valid_data <- validation(data)
  valid_size <- length(valid_data)

  train_ts <- ts(log10(train_data), frequency = 12)

  # borrowed from AirPassengers help
  arima_fit <- arima(
    train_ts,
    c(0, 1, 1),
    seasonal = list(order = c(0, 1, 1), period = 12)
  )
  raw_arima_pred <- predict(arima_fit, n.ahead = valid_size)
  arima_pred <- 10^raw_arima_pred$pred
  arima_MSE <- mean((arima_pred - valid_data)^2)

  # stl model
  stl_fit <- stlm(train_ts, s.window = 12)
  raw_stl_pred = forecast(stl_fit, h = valid_size)
  stl_pred <- 10^raw_stl_pred$mean
  stl_MSE <- mean((stl_pred - valid_data)^2)

  out <- list(
    mse = data.frame(fold = fold_index(), arima = arima_MSE, stl = stl_MSE)
  )
  return(out)
}

mses = cross_validate(
  cv_fun = cv_forecasts,
  folds = folds,
  data = AirPassengers
)$mse
colMeans(mses[, c("arima", "stl")])

## ----sessionInfo, echo=FALSE--------------------------------------------------
sessionInfo()

###############################################################################
# This example explains how to use the cross_validate function with
# parallelization using the framework of the future package.
###############################################################################

suppressMessages(library(data.table))
library(future)
data(mtcars)
set.seed(1)

# make a lot of folds
folds <- make_folds(mtcars, fold_fun = folds_bootstrap, V = 1000)

# function to calculate cross-validated squared error for linear regression
cv_lm <- function(fold, data, reg_form) {
  # get name and index of outcome variable from regression formula
  out_var <- as.character(unlist(str_split(reg_form, " "))[1])
  out_var_ind <- as.numeric(which(colnames(data) == out_var))

  # split up data into training and validation sets
  train_data <- training(data)
  valid_data <- validation(data)

  # fit linear model on training set and predict on validation set
  mod <- lm(as.formula(reg_form), data = train_data)
  preds <- predict(mod, newdata = valid_data)

  # capture results to be returned as output
  out <- list(
    coef = data.frame(t(coef(mod))),
    SE = ((preds - valid_data[, out_var_ind])^2)
  )
  return(out)
}

plan(sequential)
time_seq <- system.time({
  results_seq <- cross_validate(
    cv_fun = cv_lm, folds = folds, data = mtcars,
    reg_form = "mpg ~ ."
  )
})

plan(multicore)
time_mc <- system.time({
  results_mc <- cross_validate(
    cv_fun = cv_lm, folds = folds, data = mtcars,
    reg_form = "mpg ~ ."
  )
})

if (availableCores() > 1) {
  time_mc["elapsed"] < 1.2 * time_seq["elapsed"]
}

#' Cross-Validation Schemes
#'
#' These functions represent different cross-validation schemes that can be
#' used with \pkg{origami}. They should be used as options for the
#' \code{fold_fun} argument to \code{\link{make_folds}}, which will call the
#' requested function specify \code{n}, based on its arguments, and pass any
#' remaining arguments (e.g. \code{V} or \code{pvalidation}) on.
#'
#' @family fold generation functions
#'
#' @param n An integer indicating the number of observations.
#' @param V An integer indicating the number of folds.
#' @param pvalidation A \code{numeric} indicating the proportion of observation
#'  to be placed in the validation fold.
#' @param first_window An integer indicating the number of observations in the
#'  first training sample.
#' @param window_size An integer indicating the number of observations in each
#'  training sample.
#' @param validation_size An integer indicating the number of points in the
#'  validation samples; should be equal to the largest forecast horizon.
#' @param gap An integer indicating the number of points not included in the
#'  training or validation samples. The default is zero.
#' @param batch An integer indicating increases in the number of time points
#'  added to the training set in each iteration of cross-validation. Applicable
#'  for larger time-series. The default is one.
#' @param t An integer indicating the total amount of time to consider per
#'  time-series sample.
#' @param time An optional vector of integers of time points observed for each
#'  subject in the sample.
#' @param id An optional vector of unique identifiers corresponding to the time
#'  vector. These can be used to subset the time vector.
#'
#' @importFrom assertthat assert_that
#'
#' @return A list of \code{Fold}s.
#' @name fold_funs
NULL

###############################################################################

#' @rdname fold_funs
#' @export
folds_vfold <- function(n, V = 10L) {
  if (n <= V) {
    warning(paste0(
      "n (the number of units, clusters, or clusters in a specific strata) is ",
      n, " and V is ", V, ", so using leave-one-out CV, i.e. setting V = n"
    ))
    return(folds_loo(n))
  }
  folds <- rep(x = seq_len(V), length.out = n)

  # shuffle folds
  folds <- sample(folds)

  # generate fold vectors
  folds <- lapply(seq_len(V), fold_from_foldvec, folds)
  return(folds)
}

###############################################################################

#' @rdname fold_funs
#' @export
folds_resubstitution <- function(n) {
  list(make_fold(1L, seq_len(n), seq_len(n)))
}

###############################################################################

#' @rdname fold_funs
#' @export
folds_loo <- function(n) {
  # folds are trivial here
  folds <- seq_len(n)

  # generate fold vectors
  folds <- lapply(folds, fold_from_foldvec, folds)
  return(folds)
}

###############################################################################

#' @rdname fold_funs
#' @export
folds_montecarlo <- function(n, V = 1000L, pvalidation = 0.2) {
  assertthat::assert_that(pvalidation > 0 && pvalidation < 1)

  # calculate training sample size
  ntrain <- round((1 - pvalidation) * n)
  assertthat::assert_that(ntrain > 1)

  folds <- lapply(seq_len(V), fold_montecarlo, n, ntrain, replace = FALSE)
  return(folds)
}

# make monte carlo type folds
fold_montecarlo <- function(v, n, ntrain, replace) {
  training_set <- sample(n, size = ntrain, replace = replace)
  validation_set <- setdiff(seq_len(n), training_set)
  make_fold(v, training_set, validation_set)
}

###############################################################################

#' @rdname fold_funs
#' @export
folds_bootstrap <- function(n, V = 1000L) {
  folds <- lapply(seq_len(V), fold_montecarlo, n, n, replace = TRUE)
  return(folds)
}

###############################################################################

#' @rdname fold_funs
#' @export
folds_rolling_origin <- function(n, first_window, validation_size, gap = 0L,
                                 batch = 1L) {
  last_window <- n - (validation_size + gap)
  origins <- seq.int(first_window, last_window, by = batch)

  folds <- lapply(seq_along(origins), function(i) {
    origin <- origins[i]
    make_fold(
      v = i, training_set = seq_len(origin),
      validation_set = origin + gap + (seq_len(validation_size))
    )
  })
  return(folds)
}

###############################################################################

#' @rdname fold_funs
#' @export
folds_rolling_window <- function(n, window_size, validation_size, gap = 0L,
                                 batch = 1L) {
  last_window <- n - (validation_size + gap)
  origins <- seq.int(window_size, last_window, by = batch)

  folds <- lapply(seq_along(origins), function(i) {
    origin <- origins[i]
    make_fold(
      v = i, training_set = (seq_len(window_size)) + (i * batch - batch),
      validation_set = origin + gap + (seq_len(validation_size))
    )
  })
  return(folds)
}

###############################################################################

#' @rdname fold_funs
#' @export
folds_rolling_origin_pooled <- function(n, t, id = NULL, time = NULL,
                                        first_window, validation_size,
                                        gap = 0L, batch = 1L) {
  # check compatibility of ID and time arguments
  check_id_and_time(id = id, time = time)

  # make IDs the same length as time if only one ID provided
  if (length(id) == 1L) {
    id <- rep(x = id, length.out = length(time))
  }

  # make skeleton dataset with IDs and times
  if (is.null(id) & is.null(time)) {
    dat <- cbind.data.frame(
      time = rep(seq(t), n / t),
      id = rep(seq(n / t), each = t)
    )
  } else {
    # Index times by id (allows variability in time observed for each subject)
    dat <- cbind.data.frame(time = time, id = id)
  }

  ids <- unique(dat$id)
  times <- unique(dat$time)
  message(paste("Processing", length(ids), "samples with", t, "time points."))

  # establish rolling origin forecast for time-series cross-validation
  rolling_origin_skeleton <- folds_rolling_origin(
    t, first_window, validation_size, gap, batch
  )

  folds_rolling_origin <- lapply(rolling_origin_skeleton, function(fold) {
    train_times <- training(times)
    valid_times <- validation(times)
    train_idx <- which(dat$time %in% train_times)
    valid_idx <- which(dat$time %in% valid_times)
    fold <- make_fold(fold_index(), train_idx, valid_idx)
  })
  return(folds_rolling_origin)
}

###############################################################################

#' @rdname fold_funs
#' @export
folds_rolling_window_pooled <- function(n, t, id = NULL, time = NULL,
                                        window_size, validation_size,
                                        gap = 0L, batch = 1L) {
  # check compatibility of ID and time arguments
  check_id_and_time(id = id, time = time)

  # make IDs the same length as time if only one ID provided
  if (length(id) == 1L) {
    id <- rep(x = id, length.out = length(time))
  }

  # make skeleton dataset with IDs and times
  if (is.null(id) & is.null(time)) {
    dat <- cbind.data.frame(
      time = rep(seq(t), n / t),
      id = rep(seq(n / t), each = t)
    )
  } else {
    # Index times by id (allows variability in time observed for each subject)
    dat <- cbind.data.frame(time = time, id = id)
  }

  ids <- unique(dat$id)
  times <- unique(dat$time)
  message(paste("Processing", length(ids), "samples with", t, "time points."))

  # establish rolling window forecast for time-series cross-validation
  rolling_window_skeleton <- folds_rolling_window(
    t, window_size,
    validation_size, gap, batch
  )

  folds_rolling_window <- lapply(rolling_window_skeleton, function(fold) {
    train_times <- training(times)
    valid_times <- validation(times)
    train_idx <- which(dat$time %in% train_times)
    valid_idx <- which(dat$time %in% valid_times)
    fold <- make_fold(fold_index(), train_idx, valid_idx)
  })
  return(folds_rolling_window)
}

###############################################################################

#' @rdname fold_funs
#' @export
folds_vfold_rolling_origin_pooled <- function(n, t, id = NULL, time = NULL,
                                              V = 10L, first_window,
                                              validation_size, gap = 0L,
                                              batch = 1L) {
  # check compatibility of ID and time arguments
  check_id_and_time(id = id, time = time)

  # make IDs the same length as time if only one ID provided
  if (length(id) == 1L) {
    id <- rep(x = id, length.out = length(time))
  }

  # make skeleton dataset with IDs and times
  if (is.null(id) & is.null(time)) {
    dat <- cbind.data.frame(
      time = rep(seq(t), n / t),
      id = rep(seq(n / t), each = t)
    )
  } else {
    # Index times by id (allows variability in time observed for each subject)
    dat <- cbind.data.frame(time = time, id = id)
  }

  ids <- unique(dat$id)
  times <- unique(dat$time)
  message(paste("Processing", length(ids), "samples with", t, "time points."))

  # establish V folds for cross-validating ids
  Vfold_allocation <- sample(rep(x = seq_len(V), length.out = length(ids)))
  Vfolds_skeleton <- lapply(seq_len(V), fold_from_foldvec, Vfold_allocation)

  # establish rolling origin forecast for time-series cross-validation
  rolling_origin_skeleton <- folds_rolling_origin(
    t, first_window,
    validation_size, gap, batch
  )

  # Put it all together: gives V-fold and rolling structure
  Vfolds_rolling_origin_pooled <- lapply(Vfolds_skeleton, function(vfold) {

    # Fold-specific Sample Index
    vfold_train_id <- vfold$training_set
    vfold_valid_id <- vfold$validation_set
    vfold_train_idx <- which(dat$id %in% vfold_train_id)
    vfold_valid_idx <- which(dat$id %in% vfold_valid_id)

    # Time
    folds_rolling_origin <- lapply(rolling_origin_skeleton, function(tfold) {
      train_times <- training(times, fold = tfold)
      valid_times <- validation(times, fold = tfold)
      train_idx <- which(dat$time %in% train_times)
      valid_idx <- which(dat$time %in% valid_times)
      fold_train_idx <- which(train_idx %in% vfold_train_idx)
      fold_valid_idx <- which(valid_idx %in% vfold_valid_idx)
      fold_idx <- length(rolling_origin_skeleton) * (vfold$v - 1L) + tfold$v
      make_fold(fold_idx, fold_train_idx, fold_valid_idx)
    })
  })

  folds <- unlist(Vfolds_rolling_origin_pooled, recursive = FALSE)
  return(folds)
}

###############################################################################

#' @rdname fold_funs
#' @export
folds_vfold_rolling_window_pooled <- function(n, t, id = NULL, time = NULL,
                                              V = 10L, window_size,
                                              validation_size, gap = 0L,
                                              batch = 1L) {
  # check compatibility of ID and time arguments
  check_id_and_time(id = id, time = time)

  # make IDs the same length as time if only one ID provided
  if (length(id) == 1) {
    id <- rep(x = id, length.out = length(time))
  }

  # make skeleton dataset with IDs and times
  if (is.null(id) & is.null(time)) {
    dat <- cbind.data.frame(
      time = rep(seq(t), n / t),
      id = rep(seq(n / t), each = t)
    )
  } else {
    # Index times by id (allows variability in time observed for each subject)
    dat <- cbind.data.frame(time = time, id = id)
  }

  ids <- unique(dat$id)
  times <- unique(dat$time)
  message(paste("Processing", length(ids), "samples with", t, "time points."))

  # establish V folds for cross-validating ids
  Vfold_allocation <- sample(rep(x = seq_len(V), length.out = length(ids)))
  Vfolds_skeleton <- lapply(seq_len(V), fold_from_foldvec, Vfold_allocation)

  # establish rolling origin forecast for time-series cross-validation
  rolling_window_skeleton <- folds_rolling_window(
    t, window_size,
    validation_size, gap, batch
  )

  # Put it all together: gives V-fold and rolling structure
  Vfolds_rolling_window_pooled <- lapply(Vfolds_skeleton, function(vfold) {
    # Fold-specific Sample Index
    vfold_train_id <- vfold$training_set
    vfold_valid_id <- vfold$validation_set
    vfold_train_idx <- which(dat$id %in% vfold_train_id)
    vfold_valid_idx <- which(dat$id %in% vfold_valid_id)

    # Time
    folds_rolling_origin <- lapply(rolling_window_skeleton, function(tfold) {
      train_times <- training(times, fold = tfold)
      valid_times <- validation(times, fold = tfold)
      train_idx <- which(dat$time %in% train_times)
      valid_idx <- which(dat$time %in% valid_times)
      fold_train_idx <- which(train_idx %in% vfold_train_idx)
      fold_valid_idx <- which(valid_idx %in% vfold_valid_idx)
      fold_idx <- length(rolling_window_skeleton) * (vfold$v - 1L) + tfold$v
      make_fold(fold_idx, fold_train_idx, fold_valid_idx)
    })
  })

  folds <- unlist(Vfolds_rolling_window_pooled, recursive = FALSE)
  return(folds)
}

###############################################################################

#' Check ID and Time Compatibility
#'
#' @param id An optional vector of unique identifiers corresponding to the time
#'  vector. These can be used to subset the time vector.
#' @param time An optional vector of integers of time points observed for each
#'  subject in the sample.
#'
#' @importFrom assertthat assert_that
#'
#' @keywords internal
check_id_and_time <- function(id, time) {
  # check that both time and ID are provided, or that neither are provided
  msg_time_id <- paste(
    "Cannot create flexible folds (allow for variability",
    "in the amount of time observed for each id) unless",
    "both `time` and `id` arguments are provided. Either",
    "provide both `time` and `id` or neither."
  )
  assertthat::assert_that(!(!is.null(id) & is.null(time)) ||
    !(is.null(id) & !is.null(time)),
  msg = msg_time_id
  )

  # check that observed times are provided for each ID
  msg_time_length <- paste(
    "Cannot create flexible folds (allow for",
    "variability in the amount of `time` observed for",
    "each `id`) unless `time` vector is of same length",
    "as `id` vector. `time` is a vector of integers of",
    "time points observed for each subject, and `id`",
    "is a vector of unique identifiers which",
    "correspond to the time vector. The `id` vector",
    "is used to subset the `time` vector."
  )
  if (length(id) > 1) {
    assertthat::assert_that(length(id) == length(time),
      msg = msg_time_length
    )
  }
}

#' Wrap a Function in a Try Statement
#'
#' Function factory that generates versions of functions wrapped in \code{try}.
#'
#' @param fun A \code{function} to be wrapped in a \code{try} statement.
#' @param ... Additional arguments passed to the previous argument \code{fun}.
#'
#' @export
wrap_in_try <- function(fun, ...) {
  wrapped <- function(...) {
    try(
      {
        fun(...)
      },
      silent = TRUE
    )
  }
  return(wrapped)
}

################################################################################

#' Build a Fold Object from a Fold Vector
#'
#' For V-fold type cross-validation. This takes a fold vector (validation set
#' IDs) and builds a fold object for fold V.
#'
#' @family fold generation functions
#'
#' @param v An identifier of the fold in which observations fall for
#'  cross-validation.
#' @param folds A vector of the fold status for each observation for
#'  cross-validation.
#'
#' @export
fold_from_foldvec <- function(v, folds) {
  training_set <- which(folds != v)
  validation_set <- which(folds == v)
  make_fold(v, training_set, validation_set)
}

################################################################################

#' Build a Fold Vector from a Fold Object
#'
#' For V-fold type cross-validation. This takes a fold object and returns a fold
#' vector (containing the validation set IDs) for use with other tools like
#' \code{\link[glmnet]{cv.glmnet}}.
#'
#' @family fold generation functions
#'
#' @param folds A \code{fold} object as produced by \code{\link{make_folds}},
#'  from which a \code{numeric} vector of the validation set fold IDs are
#'  returned.
#'
#' @export
folds2foldvec <- function(folds) {
  fold_vec <- lapply(folds, function(fold) {
    validation()
  })
  len <- sapply(fold_vec, length)
  nums <- rep(seq_along(fold_vec), len)
  fold_index <- unlist(fold_vec)
  fold_id <- nums[order(fold_index)]
  return(fold_id)
}

################################################################################

is.fold <- function(x) {
  class(x) == "fold"
}

context("Folds")

############################ generic test should be applied to all generated fold
############################ vectors make sure training and validation sets don't
############################ overlap
find_overlap <- function(fold) {
  training_idx <- training()
  validation_idx <- validation()
  overlap <- intersect(training_idx, validation_idx)
  list(overlap = overlap)
}

test_splits <- function(folds) {
  overlaps <- cross_validate(find_overlap, folds)
  all_overlaps <- unlist(overlaps)
  test_that("Training and Validation don't overlap", expect_length(
    all_overlaps,
    0
  ))
}

############################

# generate v-fold fold vector
n <- 1000
folds <- make_folds(n = n, fold_fun = folds_vfold)

test_splits(folds)

# make sure v fold validation sets are exhaustive, mutually
# exclusive
get_validation_sets <- function(fold) {
  list(fold_data = data.table(validation = validation(), fold = fold_index()))
}

validation_sets <- cross_validate(get_validation_sets, folds)
validation_dt <- validation_sets$fold_data
all_indicies <- seq_len(n)
test_that("V-fold validation sets are exhaustive", {
  expect_equivalent(sort(validation_dt$validation), all_indicies)
})

max_index_count <- max(table(validation_dt$validation))
test_that("V-fold validation sets are mutually exclusive", {
  expect_equal(max_index_count, 1)
})

############################

# make sure ids all get put in the same validation set

# generate 100 subjects, each with n/100 replicates
ids <- sample(seq_len(100), n, replace = T)

id_folds <- make_folds(fold_fun = folds_vfold, cluster_id = ids)
test_splits(id_folds)

get_validation_sets_ids <- function(fold, ids) {
  list(fold_data = data.table(
    validation = validation(), fold = fold_index(),
    id = validation(ids)
  ))
}

validation_sets <- cross_validate(
  get_validation_sets_ids, id_folds,
  ids
)
validation_dt <- validation_sets$fold_data

idtab <- table(validation_dt$id, validation_dt$fold)
fold_counts <- rowSums(idtab > 0)
max_fold_count <- max(fold_counts)
test_that("Each ID only appears in one fold", {
  expect_equal(max_fold_count, 1)
})

############################

# make sure folds are roughly balanced in strata

# generate two strata, one much more rare than the other
set.seed(1)
strata_ids <- rbinom(n, 1, 0.05)

strata_folds <- make_folds(fold_fun = folds_vfold, strata_id = strata_ids)
nfolds <- length(strata_folds)
test_splits(strata_folds)


validation_sets <- cross_validate(
  get_validation_sets_ids, strata_folds,
  strata_ids
)
validation_dt <- validation_sets$fold_data
one_counts <- validation_dt[, list(one_count = sum(strata_ids)),
  by = list(fold)
]
count_range <- diff(range(one_counts$one_count))
test_that("Strata are roughly balanced", {
  expect_lte(count_range, nfolds)
})

############################

# check ids nested in strata is enforced
test_that("Ids must be nested in strata", expect_error(make_folds(
  cluster_ids = ids,
  strata_ids = strata_ids
)))

############################

# v fold->LOO fallback
smalln <- 5
suppressWarnings({
  folds <- make_folds(n = smalln, fold_fun = folds_vfold, V = 10)
})
test_splits(folds)

test_that("V fold falls back to LOO for small n, large V", {
  expect_length(folds, smalln)
})

############################

# Error if no way to guess n
test_that("Error if we can't guess n", {
  expect_error(make_folds())
})

############################

# Simple bootstrap fold test
folds <- make_folds(n = n, fold_fun = folds_bootstrap, V = 10)
test_splits(folds)

############################
# tests when number of observations in a strata < V
test_that("Reorder strata order and set V same for all strata", {
  strata_ids <- c(rep(0, 15), rep(1, 9))
  expect_warning(make_folds(strata_ids = strata_ids))
})

test_that("Reorder strata order and set V same for all strata with clusters", {
  cluster_ids <- c(1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4)
  strata_ids <- c(1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0)
  expect_warning(make_folds(cluster_ids = cluster_ids, strata_ids = strata_ids))
})

test_that("make_folds error with time series fold_fun and strata/clusters", {
  expect_error(make_folds(
    cluster_ids = c(1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4),
    strata_ids = c(1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0),
    fold_fun = folds_rolling_origin
  ))
})

test_that("folds2foldvec", {
  strata_ids <- c(rep(0, 10), rep(1, 10))
  folds <- suppressWarnings(make_folds(strata_ids = strata_ids))
  vecfolds <- folds2foldvec(folds)
  expect_equal(vecfolds, rep(seq_len(10), 2))
})

#' Fold
#'
#' Functions to make a fold. Current representation is a simple \code{list}.
#'
#' @param v An integer index of folds in the larger scheme.
#' @param training_set An integer vector of indexes corresponding to the
#'  training set.
#' @param validation_set An integer vector of indexes corresponding to the
#'  validation set.
#' @return A list containing these elements.
#'
#' @seealso \code{\link{fold_helpers}}
#'
#' @export
make_fold <- function(v, training_set, validation_set) {
  fold <- list(
    v = v,
    training_set = training_set,
    validation_set = validation_set
  )
  class(fold) <- "fold"
  return(fold)
}

# function factory for different fold-based indexing functions
get_from_fold <- function(component) {
  # function to index an object based on the training set,
  # validation set, or index of a fold
  function(x = NULL, fold = NULL) {
    # if fold isn't specified, attempt to pull from the calling
    # environment
    if (is.null(fold)) {
      if (!exists("fold", envir = parent.frame())) {
        stop("no fold specified, and fold not defined in calling function")
      }
      fold <- get("fold", envir = parent.frame())
    }
    stopifnot(is.fold(fold))

    index <- fold[[component]]

    if (is.null(x)) {
      # if no x, return indexes
      return(index)
    } else if (is.null(dim(x))) {
      # if no dim, index on only dimension
      return(x[index])
    } else {
      # if dim, index on first dimension
      return(x[index, , drop = FALSE])
    }
  }
}

################################################################################

#' Fold Helpers
#'
#' Accessors and indexers for the different parts of a fold.
#'
#' @param x an object to be indexed by a training set, validation set, or fold
#'  index. If missing, the index itself will be returned.
#' @param fold Fold; the fold used to do the indexing. If missing, \code{fold}
#'  will be pulled from the calling environment, if available.
#'
#' @return The elements of \code{x} corresponding to the indexes, or the
#'  indexes themselves if \code{x} is missing.
#'
#' @rdname fold_helpers
#'
#' @name fold_helpers
#'
#' @seealso \code{\link{make_fold}}
NULL

#' @rdname fold_helpers
#' @export
training <- get_from_fold("training_set")

#' @rdname fold_helpers
#' @export
validation <- get_from_fold("validation_set")

#' @rdname fold_helpers
#' @export
fold_index <- get_from_fold("v")

context("Fold Accessors")

test_that("Error on no fold defined", {
  expect_error(training())
})

fold <- "garbage fold"
test_that("Error on garbage fold", {
  expect_error(training())
})

training_fun <- get_from_fold("training_set")
test_that("Function factory works", {
  expect_equal(training, training_fun)
})

#' Flexible Guessing and Mapping for Combining Data Types
#'
#' Maps data types into standard combiners that should be sensible.
#'
#' @param result A single result; flexibly accepts several object classes.
#'
#' @importFrom data.table rbindlist
#'
#' @return A function to combine a list of such results.
guess_combiner <- function(result) {
  switch(class(result)[1],
    data.table = data.table::rbindlist,
    data.frame = combiner_rbind,
    matrix = combiner_rbind,
    numeric = combiner_c,
    character = combiner_c,
    integer = combiner_c,
    logical = combiner_c,
    factor = unlist,
    array = combiner_array,
    identity
  )
}

################################################################################

#' Combine Results from Different Folds
#'
#' Applies \code{\link{combiners}}: functions that collapse across a list of
#' similarly structured results, to a list of such lists.
#'
#' @param results A \code{list} of \code{list}s, corresponding to each result,
#'  with the inner \code{list}s corresponding to results from each fold.
#' @param combiners A \code{list} with the same names as results, containing
#'  combiner function names or functions for each result.
#' @param smart_combiners A \code{logical} indicating whether combiners should
#'  be guessed from the data type of the results if they are missing.
#'
#' @details In theory you should never call this function directly, because it
#'  is called automatically by cross_validate. The defaults, combiners guessed
#'  based on data type, should work in most cases.
#'
#' @return A list of combined results.
#'
#' @seealso \code{\link{combiners}}
#'
#' @export
combine_results <- function(results,
                            combiners = NULL,
                            smart_combiners = TRUE) {
  result_names <- names(results)

  if (is.null(combiners) && smart_combiners) {
    # try to learn from data structure
    first_fold <- lapply(results, function(result) result[[1]])
    combiners <- sapply(first_fold, guess_combiner)
  }

  # combine results from different folds using the appropriate combiner
  if (!is.null(combiners)) {
    combined <- lapply(result_names, function(result_name) {
      combiners[[result_name]](results[[result_name]])
    })
    names(combined) <- result_names
  } else {
    combined <- results
  }
  return(combined)
}

###############################################################################

#' Combiners
#'
#' Combiners are functions that collapse across a list of similarly structured
#' results. These are standard idioms for combining lists of certain data types.
#'
#' @param x A \code{list} of similar results to be combined.
#'
#' @return A combined results object.
#'
#' @rdname combiners
#'
#' @name combiners
NULL

#' @rdname combiners
#'
#' @export
combiner_rbind <- function(x) {
  do.call(rbind, x)
}

#' @rdname combiners
#'
#' @export
combiner_c <- function(x) {
  do.call(c, x)
}

#' @rdname combiners
#'
#' @export
combiner_factor <- function(x) {
  unlist(x)
}

#' @rdname combiners
#'
#' @importFrom abind abind
#'
#' @export
combiner_array <- function(x) {
  do.call(abind::abind, c(x, along = 1))
}

context("Combine Results")

# test setup
raw_input <- list(a = list(1, 2, 3, 4))
output <- combine_results(raw_input, combiners = NULL, smart_combiners = F)

# invoke test
test_that("combine_results falls back by returning unprocessed input", {
  expect_equal(raw_input, output)
})

# test setup
nrows <- 10
ncols <- 10
nmats <- 5
nreps <- 10
md_array <- array(0, dim = c(nrows, ncols, nmats))
array_list <- replicate(nreps, md_array, simplify = F)
combined_array <- combiner_array(array_list)
expected_dims <- c(nrows * nreps, ncols, nmats)

# invoke test
test_that("combiner_array stacks on rows for multidimensional arrays", {
  expect_equal(dim(combined_array), expected_dims)
})

#' Make List of Folds for cross-validation
#'
#' Generates a list of folds for a variety of cross-validation schemes.
#'
#' @family fold generation functions
#'
#' @param n - either an integer indicating the number of observations to
#'  cross-validate over, or an object from which to guess the number of
#'  observations; can also be computed from \code{strata_ids} or
#'  \code{cluster_ids}.
#' @param fold_fun - A function indicating the cross-validation scheme to use.
#'  See \code{\link{fold_funs}} for a list of possibilities.
#' @param cluster_ids - a vector of cluster ids. Clusters are treated as a unit
#'  -- that is, all observations within a cluster are placed in either the
#'  training or validation set.
#' @param strata_ids - a vector of strata ids. Strata are balanced: insofar as
#'  possible the distribution in the sample should be the same as the
#'  distribution in the training and validation sets.
#' @param ... other arguments to be passed to \code{fold_fun}.
#'
#' @return A list of folds objects. Each fold consists of a list with a
#'  \code{training} index vector, a \code{validation} index vector, and a
#'  \code{fold_index} (its order in the list of folds).
#'
#' @export
make_folds <- function(n = NULL,
                       fold_fun = folds_vfold,
                       cluster_ids = NULL,
                       strata_ids = NULL,
                       ...) {
  if (missing(n)) {
    # compute n from strata or cluster ids if possible
    if (!is.null(strata_ids)) {
      n <- length(strata_ids)
    } else if (!is.null(cluster_ids)) {
      n <- length(cluster_ids)
    } else {
      stop("n not specified and there are no strata or cluster IDs.")
    }
  } else if (length(n) > 1) {
    # if n not an integer, use the number of rows or length of n
    if (!is.null(nrow(n))) {
      n <- nrow(n)
    } else {
      n <- length(n)
    }
  }

  if ("t" %in% names(formals(fold_fun)) &
    (!is.null(cluster_ids) | !is.null(strata_ids))) {
    stop(
      "cluster_ids and strata_ids are not supported by fold functions ",
      "for time series cross-validation schemes. See the time series fold ",
      "function's specific arguments to check what is supported."
    )
  }

  if (!is.null(strata_ids)) {
    stopifnot(length(strata_ids) == n)

    if (!is.null(cluster_ids)) {
      stopifnot(length(cluster_ids) == n)

      # it's not clear what to do if clusters are not nested in
      # strata, so we require this for now
      nesting <- all(rowSums(table(cluster_ids, strata_ids) > 0) == 1)

      if (!nesting) {
        stop("cluster IDs are not nested in strata IDs. This is currently unsupported.")
      }
    }

    # generate separate folds for each strata
    folds <- strata_folds(fold_fun, cluster_ids, strata_ids, ...)
  } else if (!is.null(cluster_ids)) {
    # generate folds on clusters instead of observations
    stopifnot(length(cluster_ids) == n)
    folds <- cluster_folds(fold_fun, cluster_ids, ...)
  } else {
    # we either don't have clusters or strata, or we're in the
    # functions that are handling those
    # generate folds
    folds <- fold_fun(n, ...)
  }
  return(folds)
}

################################################################################

#' Convert ID Folds to Observation Folds
#'
#' This function convertsf olds that subset ids to folds that subset
#'  observations
#'
#' @param idfolds folds that subset ids
#' @param cluster_ids a vector of cluster ids indicating which observations are
#'  in which clusters
#'
#' @export
id_folds_to_folds <- function(idfolds, cluster_ids) {
  idfac <- factor(cluster_ids)
  nclusters <- length(levels(idfac))
  clusternums <- as.numeric(idfac)

  reindex <- function(index, fold_index) {
    which(index %in% fold_index)
  }

  folds <- lapply(idfolds, function(idfold) {
    make_fold(
      v = fold_index(fold = idfold),
      training_set = reindex(clusternums, training(fold = idfold)),
      validation_set = reindex(clusternums, validation(fold = idfold))
    )
  })

  return(folds)
}

# Generate folds for clusters, and then convert into folds for observations
cluster_folds <- function(fold_fun, cluster_ids, ...) {
  # convert ids to numeric 1:n
  idfac <- factor(cluster_ids)
  nclusters <- length(levels(idfac))
  clusternums <- as.numeric(idfac)
  # id_indexes <- by(seq_along(cluster_ids), list(id = clusternums), list)

  # generate folds for ids
  idfolds <- make_folds(
    n = nclusters, fold_fun = fold_fun,
    cluster_ids = NULL, ...
  )

  folds <- id_folds_to_folds(idfolds, cluster_ids)
  return(folds)
}

################################################################################

# generate folds separately for each strata, and then collapse
strata_folds <- function(fold_fun, cluster_ids, strata_ids, ...) {

  # order strata by increasing prevalence, so rarest strata is listed first
  ordered <- names(sort(prop.table(table(strata_ids)), decreasing = FALSE))
  idfac <- sort(factor(strata_ids, levels = ordered))
  nstrata <- length(levels(idfac))
  stratanums <- as.numeric(idfac)

  ##### generate strata-specific folds
  # create list args for calling make_folds
  args <- list(...)
  if (is.null(args$V) & "V" %in% names(formals(fold_fun))) {
    args$V <- formals(fold_fun)$V
  }
  args$fold_fun <- fold_fun
  args$cluster_ids <- cluster_ids
  # if number of observations in rarest strata < V, then V needs to change for
  # all strata-specific folds, so V is same across all strata-specific folds
  n_rare <- sum(stratanums == 1)
  if (!is.null(args$V) && n_rare < args$V) {
    warning(paste0(
      "The number of observations in the most rare strata is ", n_rare,
      ", and V is ", args$V, ", so using leave-one-out CV, i.e. setting V = n"
    ))
    args$V <- n_rare
  }
  strata_folds <- lapply(seq_len(nstrata), function(strata) {
    args_strata <- args
    args_strata$n <- sum(stratanums == strata)
    if (is.null(args$cluster_ids)) {
      args_strata <- args_strata[names(args_strata) %in% names(formals(fold_fun))]
      suppressWarnings(do.call(args$fold_fun, args_strata))
    } else {
      args_strata$cluster_ids <- args$cluster_ids[stratanums == strata]
      do.call(make_folds, args_strata)
    }
  })

  ###### collapse strata-specific folds
  V <- length(strata_folds[[1]])
  folds <- lapply(seq_len(V), function(v) {
    # convert to indexes on the observations
    converted_folds <- lapply(seq_len(nstrata), function(strata) {
      strata_idx <- which(stratanums == strata)
      strata_fold <- strata_folds[[strata]][[v]]
      make_fold(
        v = v, training_set = training(strata_idx, strata_fold),
        validation_set = validation(strata_idx, strata_fold)
      )
    })

    # collapse across strata
    make_fold(
      v = v,
      training_set = unlist(
        lapply(converted_folds, function(fold) training(fold = fold))
      ),
      validation_set = unlist(
        lapply(converted_folds, function(fold) validation(fold = fold))
      )
    )
  })

  return(folds)
}

################################################################################

#' Repeated Cross-Validation
#'
#' Implementation of repeated window cross-validation: generates fold objects
#' for repeated cross-validation by making repeated calls to
#' \code{\link{make_folds}} and concatenating the results.
#'
#' @family fold generation functions
#'
#' @param repeats An integer indicating the number of repeats.
#' @param ... Arguments passed to \code{\link{make_folds}}.
#'
#' @export
make_repeated_folds <- function(repeats, ...) {
  all_folds <- lapply(seq_len(repeats), function(x) make_folds(...))
  folds <- unlist(all_folds, recursive = FALSE)
  return(folds)
}

context("Overall Test for Time Series")

if (require("forecast")) {
  set.seed(1)
  data(AirPassengers)

  # simple fold
  folds <- make_folds(
    AirPassengers,
    fold_fun = folds_rolling_origin,
    first_window = 36, validation_size = 24, gap = 0, batch = 1
  )
  fold <- folds[[1]]

  # function to calculate cross-validated squared error
  cvforecasts <- function(fold) {
    train_data <- training(AirPassengers)
    valid_data <- validation(AirPassengers)
    valid_size <- length(valid_data)

    train_ts <- ts(log10(train_data), frequency = 12)

    # borrowed from AirPassengers help
    arima_fit <- arima(
      train_ts, c(0, 1, 1),
      seasonal = list(
        order = c(0, 1, 1),
        period = 12
      )
    )
    raw_arima_pred <- predict(arima_fit, n.ahead = valid_size)
    arima_pred <- 10^raw_arima_pred$pred
    arima_MSE <- mean((arima_pred - valid_data)^2)

    # stl model
    stl_fit <- stlm(train_ts, s.window = 12)
    raw_stl_pred <- forecast(stl_fit, h = valid_size)
    stl_pred <- 10^raw_stl_pred$mean
    stl_MSE <- mean((stl_pred - valid_data)^2)

    list(mse = data.frame(
      fold = fold_index(), arima = arima_MSE,
      stl = stl_MSE
    ))
  }

  mses <- cross_validate(cvforecasts, folds)$mse
  mses_mean <- colMeans(mses[, c("arima", "stl")])

  # Test we get the same result as in the vignette example:
  test_that("CV-MSE matches previous value", {
    expect_equal(mses_mean[[1]], 667.2478, tolerance = 0.01)
  })

  # Tests with gap and batch parameters:
  folds <- make_folds(
    AirPassengers,
    fold_fun = folds_rolling_origin,
    first_window = 36, validation_size = 24, gap = 5, batch = 10
  )

  fold <- folds[[1]]
  mses <- cross_validate(cvforecasts, folds)$mse
  mses_mean <- colMeans(mses[, c("arima", "stl")])

  test_that("CV-MSE with gap and batch matches previous value", {
    expect_equal(mses_mean[[1]], 6004.730, tolerance = 0.01)
  })


  # Tests with gap and batch parameters for rolling window CV:
  folds <- make_folds(
    AirPassengers,
    fold_fun = folds_rolling_window,
    window_size = 36, validation_size = 24, gap = 5, batch = 2
  )

  fold <- folds[[1]]
  mses <- cross_validate(cvforecasts, folds)$mse
  mses_mean <- colMeans(mses[, c("arima", "stl")])

  test_that("CV-MSE with rolling window matches previous value", {
    expect_equal(mses_mean[[1]], 7580.455, tolerance = 0.01)
  })

  #############################################################################
  # Test multiple time-series functionality
  n_id <- 100
  test_data <- data.table(id = seq_len(n_id))
  test_data <- test_data[, list(t = seq_len(floor(runif(1) * 50))),
    by = list(id)
  ]
  test_data[, X := rnorm(.N)]

  ### Independent sample example
  folds <- make_folds(test_data,
    fold_fun = folds_rolling_origin_pooled,
    id = test_data$id,
    t = max(test_data$t),
    time = test_data$t, first_window = 4,
    validation_size = 4, gap = 0, batch = 2
  )

  test_that("Size of the first fold of rolling origin pooled CV", {
    expect_equal(length(folds[[1]]$training_set), sum(test_data$t <= 4))
  })

  folds <- make_folds(test_data,
    fold_fun = folds_rolling_window_pooled,
    id = test_data$id,
    t = max(test_data$t),
    time = test_data$t,
    window_size = 4,
    validation_size = 4, gap = 0, batch = 2
  )
  test_that("Size of the first fold of rolling window pooled CV", {
    expect_equal(length(folds[[1]]$training_set), sum(test_data$t <= 4))
  })

  ### Independent sample example:
  #   not the same number of time points
  #   id based
  test_data_id <- data.table(melt(data.table(AirPassengers),
    measure.vars = "AirPassengers"
  ),
  id = c(rep(1, 60), rep(2, 84))
  )

  folds_id1 <- make_folds(test_data_id,
    fold_fun = folds_rolling_window_pooled,
    t = 60, window_size = 5, id = 1, time = seq(1:60),
    validation_size = 5, gap = 0, batch = 20
  )
  folds_id2 <- make_folds(test_data_id,
    fold_fun = folds_rolling_window_pooled,
    t = 84, window_size = 5, id = 2, time = as.matrix(seq(1:84)),
    validation_size = 7, gap = 0, batch = 20
  )
  test_that("Size of CVs rolling window pooled CV, id 2", {
    expect_equal(length(folds_id2), 4)
  })
  test_that("Size of CVs rolling window pooled CV, id 1", {
    expect_equal(length(folds_id1), 3)
  })

  folds_id1 <- make_folds(test_data_id,
    fold_fun = folds_rolling_origin_pooled,
    t = 60, first_window = 10, id = 1, time = seq(1:60),
    validation_size = 5, gap = 0, batch = 20
  )
  folds_id2 <- make_folds(test_data_id,
    fold_fun = folds_rolling_origin_pooled,
    t = 84, first_window = 10, id = 2, time = seq(1:84),
    validation_size = 5, gap = 0, batch = 20
  )
  test_that("Size of CVs rolling origin pooled CV, id 2", {
    expect_equal(length(folds_id2), 4)
  })
  test_that("Size of CVs rolling origin pooled CV, id 1", {
    expect_equal(length(folds_id1), 3)
  })

  ### Dependent samples example
  folds <- make_folds(test_data,
    fold_fun = folds_vfold_rolling_origin_pooled,
    t = 12, first_window = 6, V = 5,
    validation_size = 2, gap = 0, batch = 2
  )
  test_that("Dimension of folds for the V-fold rolling origin pooled CV", {
    expect_equal(length(folds), 15)
  })

  folds <- make_folds(test_data,
    fold_fun = folds_vfold_rolling_window_pooled,
    t = 12, window_size = 6, V = 5,
    validation_size = 2, gap = 0, batch = 3
  )
  test_that("Dimension of folds for the V-fold rolling window pooled CV", {
    expect_equal(length(folds), 10)
  })

  ### Example with multivariate time series with variations in what's observed:
  # for each id, vary time measured and number of observations
  value <- AirPassengers
  time <- c(1:48, c(1, 5:51), c(1:10, 31:68))
  id <- rep(1:3, each = 48)
  dat <- data.table(id, time, value)

  folds <- make_folds(dat,
    t = 40, id = dat$id, time = dat$time,
    fold_fun = folds_rolling_origin_pooled,
    first_window = 10, validation_size = 10, gap = 0, batch = 10
  )

  # rows which should not be included in any folds
  bad_index1 <- which(dat$time > 40)
  test_that("Time included in folds does not exceed t", {
    expect_false(any(bad_index1 %in% unlist(folds)))
  })

  # no rows from id 3 should be included in fold 2 validation set
  # since time 21-30 not observed for id 3, and fold 2 validation is times 21-30
  id3 <- which(dat$id == 3)
  test_that("Validation folds respect differences in time for each id", {
    expect_false(any(id3 %in% folds[[2]]$validation_set))
  })

  # only 7 rows from id 2 should be included in fold 1 training set
  # since times 2-4 not observed for id 2, and fold 1 training is times 1-10
  id2_fold1 <- which(dat$id == 2 & dat$time <= 10)
  test_that("Training folds respect differences in time for each id", {
    expect_equal(length(id2_fold1), 7)
  })
}

context("Main Cross-Validation Method")

# test error handling for failed folds
cv_sometimes_errors <- function(fold) {
  index <- fold_index()
  if (index == 7) {
    stop("7 makes me sad")
  }


  return(list(a = "certainly not 7"))
}

folds <- make_folds(1000)
results <- cross_validate(cv_sometimes_errors, folds)

test_that("Errors are put in special error vector", {
  expect_length(results$errors$error, 1)
})


# test error handling when all folds fail
cv_always_errors <- function(fold) {
  stop("I always produce an error")
}

folds <- make_folds(1000)
suppressWarnings({
  results <- cross_validate(cv_always_errors, folds)
})

test_that("cross_validate handles the case where all folds produce an error", {
  expect_equal(names(results), "errors")
})

library(origami)
library(data.table)
context("Overall Test")

r_major <- as.numeric(R.Version()$major)
r_minor <- as.numeric(R.Version()$minor)
if (r_major == 3 && r_minor < 6) {
  cvmse_result <- 13.32
} else {
  cvmse_result <- 14.11
}
set.seed(1)
data(mtcars)

# re-substitution MSE
r <- lm(mpg ~ ., data = mtcars)
simple_resub_MSE <- mean(resid(r)^2)

# function to calculate cross-validated squared error
cvlm <- function(fold) {
  train_data <- training(mtcars)
  valid_data <- validation(mtcars)

  r <- lm(mpg ~ ., data = train_data)
  preds <- predict(r, newdata = valid_data)
  list(coef = data.frame(t(coef(r))), SE = ((preds - valid_data$mpg)^2))
}

# replicate the resubstitution estimate
resub <- make_folds(mtcars, fold_fun = folds_resubstitution)[[1]]
resub_results <- cvlm(resub)
resub_MSE <- mean(resub_results$SE)
test_that("Resub MSE matches simple resub MSE", {
  expect_equal(resub_MSE, simple_resub_MSE, tolerance = 0.01)
})

# cross-validated estimate
folds <- make_folds(mtcars)
results <- cross_validate(cvlm, folds)
cv_MSE <- mean(results$SE)
test_that("CV MSE matches previous value", {
  expect_equal(cv_MSE, cvmse_result, tolerance = 0.01)
})

library(future)
context("Future Plan")

set.seed(94720)

data(mtcars)
# make a lot of folds
folds <- make_folds(mtcars, fold_fun = folds_bootstrap, V = 500L)

# function to calculate cross-validated squared error
cvlm <- function(fold) {
  train_data <- training(mtcars)
  valid_data <- validation(mtcars)

  r <- lm(mpg ~ ., data = train_data)
  suppressWarnings(
    preds <- predict(r, newdata = valid_data)
  )
  list(coef = data.frame(t(coef(r))), SE = ((preds - valid_data$mpg)^2))
}

plan(sequential)
time_seq <- system.time({
  results_seq <- cross_validate(cvlm, folds)
})

plan(multicore)
time_mc <- system.time({
  results_mc <- cross_validate(cvlm, folds)
})

if (future::availableCores() > 1L) {
  test_that("MC is not significantly slower than sequential", {
    skip_on_os("windows") # Windows doesn't support multicore
    expect_lt(time_mc["elapsed"], 1.2 * time_seq["elapsed"])
  })
}

######

# # verify globals are being transferred to other sessions
# a=4
# folds=make_folds(1000)
# return_a=function(fold){
#     list(a=a,dt=data.table(a=a))
# }

# plan(multisession,workers=2)
# results=cross_validate(return_a,folds)
# test_that("globals are available in worker sessions",expect_true(all(results$a==4)))
# test_that("globals are available in worker sessions",expect_is(results$dt,"data.table"))

context("cv_fun handles argument 'x' but not 'X' due to lapply conventions.")

# returns 'nrow' of input x
nrow_x_1 <- function(fold, x) {
  train_x <- training(x)
  valid_x <- validation(x)
  out <- nrow(valid_x)
  return(list(nrow_valid = out))
}

# returns 'nrow' of input X (note: 'X' is reserved in lapply / future_lapply)
nrow_X_2 <- function(fold, X) {
  train_X <- training(X)
  valid_X <- validation(X)
  out <- nrow(valid_X)
  return(list(nrow_valid = out))
}

# create fold object for using origami::cross_validate
folds <- make_folds(mtcars)

# cross_validate behaves as expected with cv_fun naming input argument 'x'
test_that("cross_validate handles cv_fun with input argument 'x'", {
  expect_equal(
    cross_validate(
      cv_fun = nrow_x_1, folds = folds,
      x = mtcars
    )$nrow_valid,
    c(4, 4, 3, 3, 3, 3, 3, 3, 3, 3)
  )
})

# cross_validate fails with cv_fun naming input argument 'X'
test_that("cross_validate rejects cv_fun with input argument 'X'", {
  expect_error(
    cross_validate(cv_fun = nrow_X_2, folds = folds, X = mtcars),
    "Argument name 'X' not allowed in cv_fun",
    ignore.case = TRUE
  )
})

#' Main Cross-Validation Function
#'
#' Applies \code{cv_fun} to the folds using \code{future_lapply} and combines
#' the results across folds using \code{\link{combine_results}}.
#'
#' @param cv_fun A function that takes a 'fold' as it's first argument and
#'  returns a list of results from that fold. NOTE: the use of an argument
#'  named 'X' is specifically disallowed in any input function for compliance
#'  with the functions \code{\link[future.apply]{future_lapply}} and
#'  \code{lapply}.
#' @param folds A list of folds to loop over generated using
#'  \code{\link{make_folds}}.
#' @param ... Other arguments passed to \code{cvfun}.
#' @param use_future A \code{logical} option for whether to run the main loop
#'  of cross-validation with \code{\link[future.apply]{future_lapply}} or with
#'  \code{lapply}.
#' @param .combine A \code{logical} indicating if \code{\link{combine_results}}
#'  should be called.
#' @param .combine_control A \code{list} of arguments to
#'  \code{\link{combine_results}}.
#' @param .old_results A \code{list} containing the returned result from a
#'  previous call to this function. Will be combined with the current results.
#'  This is useful for adding additional CV folds to a results object.
#'
#' @importFrom assertthat assert_that
#' @importFrom future future plan sequential multicore multisession
#' @importFrom future.apply future_lapply
#' @importFrom listenv listenv
#' @import methods
#'
#' @return A \code{list} of results, combined across folds.
#'
#' @export
#'
#' @example /inst/examples/cv_simple_example.R
#' @example /inst/examples/cv_parallel_example.R
cross_validate <- function(cv_fun,
                           folds,
                           ...,
                           use_future = TRUE,
                           .combine = TRUE,
                           .combine_control = list(),
                           .old_results = NULL) {

  # argument 'X' is specifically disallowed since reserved by lapply family
  args_cv_fun <- names(formals(cv_fun))
  assertthat::assert_that(!("X" %in% args_cv_fun),
    msg = "Argument name 'X' not allowed in cv_fun"
  )

  # catch and return errors if they occur
  wrapped_fun <- wrap_in_try(cv_fun)

  # main loop
  if (use_future) {
    results <- future.apply::future_lapply(folds, wrapped_fun, ...,
      future.seed = TRUE
    )
  } else {
    results <- lapply(folds, wrapped_fun, ...)
  }

  # remove error results
  error_idx <- which(sapply(results, inherits, "try-error"))
  error_results <- list(index = error_idx, error = results[error_idx])
  good_results <- setdiff(seq_along(folds), error_idx)
  results <- results[good_results]

  if (length(results) > 0) {
    # verify that the folds returned similar results
    if (length(unique(lapply(results, length))) > 1) {
      stop("lists returned from folds are not the same length")
    }
    if (length(unique(lapply(results, names))) > 1) {
      stop("names returned from folds are not consistent")
    }

    # invert results - go from a list containing one list per fold to a list
    # containing one list per result returned by cv_fun
    results <- apply(do.call(rbind, results), 2, as.list)

    # combine results
    if (.combine) {
      results <- do.call(
        combine_results,
        c(
          list(results = results),
          .combine_control
        )
      )
    }
  } else {
    warning("All iterations resulted in errors")
  }

  results$errors <- error_results

  if (!is.null(.old_results)) {
    # invert results - go from a list containing one list per fold to a list
    # containing one list per result returned by cv_fun
    new_and_old <- list(results, .old_results)
    new_and_old <- apply(do.call(rbind, new_and_old), 2, as.list)
    results <- combine_results(results = new_and_old)
  }
  return(results)
}

---
title: "Generalized Cross-Validation with Origami"
author: "Jeremy Coyle & Nima Hejazi"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
bibliography: refs.bib
vignette: >
  %\VignetteIndexEntry{Generalized Cross-Validation with Origami}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

## Introduction

Cross-validation is an essential tool for evaluating how any given data analytic
procedure extends from a sample to the target population from which the sample
is derived. It has seen widespread application in all facets of statistics,
perhaps most notably statistical machine learning.  When used for model
selection, cross-validation has powerful optimality properties
[@vaart2006oracle], [@vdl2007super].

<!--
and for CV-TMLE based parameter estimation, elimination of difficult-to-assess
empirical process conditions from CV-TMLE step [@Zheng:2010ua].
-->

Cross-validation works by partitioning a sample into complementary subsets,
applying a particular data analytic (statistical) routine on a subset (the
"training" set), and evaluating the routine of choice on the complementary
subset (the "testing" set). This procedure is repeated across multiple
partitions of the data. A variety of different  partitioning schemes exist, such
as V-fold cross-validation and bootstrap cross-validation, many of which are
supported by `origami`. The `origami` package provides a suite of tools that
generalize  the application of cross-validation to arbitrary data analytic
procedures. The use of `origami` is best illustrated by example.

---

## Cross-validation with linear regression

We'll start by examining a fairly simple data set:
```{r load_data}
data(mtcars)
head(mtcars)
```

One might be interested in examining how the efficiency of a car, as measured by
miles-per-gallon (mpg), is explained by various technical aspects of the car,
with data across a variety of different models of cars. Linear regression is
perhaps the simplest statistical procedure that could be used to make such
deductions. Let's try it out:
```{r linear_mod}
lm_mod <- lm(mpg ~ ., data = mtcars)
summary(lm_mod)
```

We can assess how well the model fits the data by comparing the predictions of
the linear model to the true outcomes observed in the data set. This is the
well known (and standard) mean squared error. We can extract that from the `lm` 
model object like so:

```{r get_naive_error}
err <- mean(resid(lm_mod)^2)
```

The mean squared error is `r err`. There is an important problem that arises
when we assess the model in this way -- that is, we have trained our linear
regression model on the full data set and assessed the error on the full data
set, using up all of our data. We, of course, are generally not interested in
how well the model explains variation in the observed data; rather, we are
interested in how the explanation provided by the model generalizes to a target
population from which the sample is presumably derived. Having used all of our
available data, we cannot honestly evaluate how well the model fits (and thus
explains) variation at the population level.

To resolve this issue, cross-validation allows for a particular procedure (e.g.,
linear regression) to be implemented over subsets of the data, evaluating how
well the procedure fits on a testing ("validation") set, thereby providing an
honest evaluation of the error.

We can easily add cross-validation to our linear regression procedure using
`origami`. First, let us define a new function to perform linear regression on a
specific partition of the data (called a "fold"):

```{r define_fun_cv_lm}
cv_lm <- function(fold, data, reg_form) {
  # get name and index of outcome variable from regression formula
  out_var <- as.character(unlist(str_split(reg_form, " "))[1])
  out_var_ind <- as.numeric(which(colnames(data) == out_var))

  # split up data into training and validation sets
  train_data <- training(data)
  valid_data <- validation(data)

  # fit linear model on training set and predict on validation set
  mod <- lm(as.formula(reg_form), data = train_data)
  preds <- predict(mod, newdata = valid_data)

  # capture results to be returned as output
  out <- list(
    coef = data.frame(t(coef(mod))),
    SE = ((preds - valid_data[, out_var_ind])^2)
  )
  return(out)
}
```

Our `cv_lm` function is rather simple: we merely split the available data into a
training and validation sets, using the eponymous functions provided in
`origami`, fit the linear model on the training set, and evaluate the model on
the testing set. This is a simple example of what `origami` considers to be
`cv_fun`s -- functions for using cross-validation to perform a particular
routine over an input data set. Having defined such a function, we can simply
generate a set of partitions using `origami`'s `make_folds` function, and apply
our `cv_lm` function over the resultant `folds` object. Below, we replicate the
resubstitution estimate of the error -- we did this "by hand" above -- using
the functions `make_folds` and `cv_lm`.

```{r load_pkgs}
library(origami)
library(stringr) # used in defining the cv_lm function above
```

```{r cv_lm_resub}
# resubstitution estimate
resub <- make_folds(mtcars, fold_fun = folds_resubstitution)[[1]]
resub_results <- cv_lm(fold = resub, data = mtcars, reg_form = "mpg ~ .")
mean(resub_results$SE)
```

This (very nearly) matches the estimate of the error that we obtained above.

We can more honestly evaluate the error by _V-fold cross-validation_, which
partitions the data into __v subsets__, fitting the model on $v - 1$ of the
subsets and evaluating on the subset that was held out for testing. This is
repeated such that each subset is used for testing. We can easily apply our
`cv_lm` function using `origami`'s `cross_validate` (n.b., by default this
performs 10-fold cross-validation):

```{r cv_lm_cross_valdate}
# cross-validated estimate
folds <- make_folds(mtcars)
cvlm_results <- cross_validate(
  cv_fun = cv_lm,
  folds = folds,
  data = mtcars,
  reg_form = "mpg ~ ."
)
mean(cvlm_results$SE)
```

Having performed 10-fold cross-validation, we quickly notice that our previous
estimate of the model error (by resubstitution) was quite optimistic. The honest
estimate of the error is several times larger.

---

## General workflow

Generally, `cross_validate` usage will mirror the workflow in the above example.
First, the user must define folds and a function that operates on each fold.
Once these are passed to `cross_validate`, the function will map the function
across the folds, and combine the results in a reasonable way. More details on
each step of this process will be given below.

### Define folds

The `folds` object passed to `cross_validate` is a list of folds. Such lists can
be generated using the `make_folds` function. Each `fold` consists of a list
with a `training` index vector, a `validation` index vector, and a `fold_index`
(its order in the list of folds). This function supports a variety of
cross-validation schemes including _v-fold_ and _bootstrap_ cross-validation as
well as time series methods like _"Rolling Window"_. It can balance across levels of a
variable (`stratify_ids`), and it can also keep all observations from the same
independent unit together (`cluster_ids`). See the documentation of the `make_folds`
function for details about supported cross-validation schemes and arguments.

### Define fold function

The `cv_fun` argument to `cross_validate` is a function that will perform some
operation on each fold. The first argument to this function must be `fold`,
which will receive an individual fold object to operate on. Additional arguments
can be passed to `cv_fun` using the `...` argument to `cross_validate`. Within
this function, the convenience functions `training`, `validation` and
`fold_index` can return the various components of a fold object. They do this by
retrieving the `fold` object from their calling environment. It can also be
specified directly. If `training` or `validation` is passed an object, it will
index into it in a sensible way. For instance, if it is a vector, it will index
the vector directly. If it is a `data.frame` or `matrix`, it will index rows.
This allows the user to easily partition data into training and validation sets.
This fold function must return a named list of results containing whatever
fold-specific outputs are generated.

### Apply `cross_validate`

After defining folds, `cross_validate` can be used to map the `cv_fun` across
the `folds` using `future_lapply`. This means that it can be easily parallelized
by specifying a parallelization scheme (i.e., a `plan`). See the [`future`
package](https://github.com/futureverse/future) for more details.

The application of `cross_validate` generates a list of results. As described
above, each call to `cv_fun` itself returns a list of results, with different
elements for each type of result we care about. The main loop generates a list
of these individual lists of results (a sort of "meta-list"). This "meta-list"
is then inverted such that there is one element per result type (this too is a
list of the results for each fold). By default, `combine_results` is used to
combine these results type lists.

For instance, in the above `mtcars` example, the results type lists contains one
`coef` `data.frame` from each fold. These are `rbind`ed together to form one
`data.frame` containing the `coefs` from all folds in different rows. How
results are combined is determined automatically by examining the data types
of the results from the first fold. This can be modified by specifying a list of
arguments to `.combine_control`. See the help for `combine_results` for more
details. In most cases, the defaults should suffice.

---

## Cross-validation with random forests

To examine `origami` further, let us return to our example analysis using the
`mtcars` data set. Here, we will write a new `cv_fun` type object. As an
example, we will use L. Breiman's `randomForest`:

```{r cv_fun_randomForest}
cv_rf <- function(fold, data, reg_form) {
  # get name and index of outcome variable from regression formula
  out_var <- as.character(unlist(str_split(reg_form, " "))[1])
  out_var_ind <- as.numeric(which(colnames(data) == out_var))

  # define training and validation sets based on input object of class "folds"
  train_data <- training(data)
  valid_data <- validation(data)

  # fit Random Forest regression on training set and predict on holdout set
  mod <- randomForest(formula = as.formula(reg_form), data = train_data)
  preds <- predict(mod, newdata = valid_data)

  # define output object to be returned as list (for flexibility)
  out <- list(
    coef = data.frame(mod$coefs),
    SE = ((preds - valid_data[, out_var_ind])^2)
  )
  return(out)
}
```

Above, in writing our `cv_rf` function to cross-validate `randomForest`, we used
our previous function `cv_lm` as an example. For now, individual `cv_fun`s must
be written by hand; however, in future releases, a wrapper may be available to
support auto-generating `cv_fun`s to be used with `origami`.

Below, we use `cross_validate` to apply our new `cv_rf` function over the `folds`
object generated by `make_folds`.

```{r}
library(randomForest)
folds <- make_folds(mtcars)
cvrf_results <- cross_validate(
  cv_fun = cv_rf,
  folds = folds,
  data = mtcars,
  reg_form = "mpg ~ ."
)
mean(cvrf_results$SE)
```

Using 10-fold cross-validation (the default), we obtain an honest estimate of
the prediction error of random forests. From this, we gather that the use of
`origami`'s `cross_validate` procedure can be generalized to arbitrary esimation
techniques, given availability of an appropriate `cv_fun` function.

---

## Cross-validation with dependence: time series

Cross-validation can also be used for forecast model selection in a time series
setting. Here, the partitioning scheme mirrors the application of the
forecasting model: We'll train the data on past observations (either all
available or a recent subset), and then use the model forecast (predict), the
next few observations. Consider the `AirPassengers` dataset, a monthly time
series of passenger air traffic in thousands of people.

```{r}
data(AirPassengers)
print(AirPassengers)
```

Suppose we want to pick between two forecasting models, `stl`, and `arima` (the
details of these models are not important for this example). We can do that by
evaluating their forecasting performance.

```{r}
library(forecast)
folds = make_folds(
  AirPassengers,
  fold_fun = folds_rolling_origin,
  first_window = 36,
  validation_size = 24
)
fold = folds[[1]]

# function to calculate cross-validated squared error
cv_forecasts <- function(fold, data) {
  train_data <- training(data)
  valid_data <- validation(data)
  valid_size <- length(valid_data)

  train_ts <- ts(log10(train_data), frequency = 12)

  # borrowed from AirPassengers help
  arima_fit <- arima(
    train_ts,
    c(0, 1, 1),
    seasonal = list(order = c(0, 1, 1), period = 12)
  )
  raw_arima_pred <- predict(arima_fit, n.ahead = valid_size)
  arima_pred <- 10^raw_arima_pred$pred
  arima_MSE <- mean((arima_pred - valid_data)^2)

  # stl model
  stl_fit <- stlm(train_ts, s.window = 12)
  raw_stl_pred = forecast(stl_fit, h = valid_size)
  stl_pred <- 10^raw_stl_pred$mean
  stl_MSE <- mean((stl_pred - valid_data)^2)

  out <- list(
    mse = data.frame(fold = fold_index(), arima = arima_MSE, stl = stl_MSE)
  )
  return(out)
}

mses = cross_validate(
  cv_fun = cv_forecasts,
  folds = folds,
  data = AirPassengers
)$mse
colMeans(mses[, c("arima", "stl")])
```

---

## Session Information

```{r sessionInfo, echo=FALSE}
sessionInfo()
```

---

## References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cross_validate.R
\name{cross_validate}
\alias{cross_validate}
\title{Main Cross-Validation Function}
\usage{
cross_validate(
  cv_fun,
  folds,
  ...,
  use_future = TRUE,
  .combine = TRUE,
  .combine_control = list(),
  .old_results = NULL
)
}
\arguments{
\item{cv_fun}{A function that takes a 'fold' as it's first argument and
returns a list of results from that fold. NOTE: the use of an argument
named 'X' is specifically disallowed in any input function for compliance
with the functions \code{\link[future.apply]{future_lapply}} and
\code{lapply}.}

\item{folds}{A list of folds to loop over generated using
\code{\link{make_folds}}.}

\item{...}{Other arguments passed to \code{cvfun}.}

\item{use_future}{A \code{logical} option for whether to run the main loop
of cross-validation with \code{\link[future.apply]{future_lapply}} or with
\code{lapply}.}

\item{.combine}{A \code{logical} indicating if \code{\link{combine_results}}
should be called.}

\item{.combine_control}{A \code{list} of arguments to
\code{\link{combine_results}}.}

\item{.old_results}{A \code{list} containing the returned result from a
previous call to this function. Will be combined with the current results.
This is useful for adding additional CV folds to a results object.}
}
\value{
A \code{list} of results, combined across folds.
}
\description{
Applies \code{cv_fun} to the folds using \code{future_lapply} and combines
the results across folds using \code{\link{combine_results}}.
}
\examples{
###############################################################################
# This example explains how to use the cross_validate function naively.
###############################################################################
data(mtcars)

# resubstitution MSE
r <- lm(mpg ~ ., data = mtcars)
mean(resid(r)^2)

# function to calculate cross-validated squared error
cv_lm <- function(fold, data, reg_form) {
  # get name and index of outcome variable from regression formula
  out_var <- as.character(unlist(stringr::str_split(reg_form, " "))[1])
  out_var_ind <- as.numeric(which(colnames(data) == out_var))

  # split up data into training and validation sets
  train_data <- training(data)
  valid_data <- validation(data)

  # fit linear model on training set and predict on validation set
  mod <- lm(as.formula(reg_form), data = train_data)
  preds <- predict(mod, newdata = valid_data)

  # capture results to be returned as output
  out <- list(
    coef = data.frame(t(coef(mod))),
    SE = ((preds - valid_data[, out_var_ind])^2)
  )
  return(out)
}

# replicate the resubstitution estimate
resub <- make_folds(mtcars, fold_fun = folds_resubstitution)[[1]]
resub_results <- cv_lm(fold = resub, data = mtcars, reg_form = "mpg ~ .")
mean(resub_results$SE)

# cross-validated estimate
folds <- make_folds(mtcars)
cv_results <- cross_validate(
  cv_fun = cv_lm, folds = folds, data = mtcars,
  reg_form = "mpg ~ ."
)
mean(cv_results$SE)
###############################################################################
# This example explains how to use the cross_validate function with
# parallelization using the framework of the future package.
###############################################################################

suppressMessages(library(data.table))
library(future)
data(mtcars)
set.seed(1)

# make a lot of folds
folds <- make_folds(mtcars, fold_fun = folds_bootstrap, V = 1000)

# function to calculate cross-validated squared error for linear regression
cv_lm <- function(fold, data, reg_form) {
  # get name and index of outcome variable from regression formula
  out_var <- as.character(unlist(str_split(reg_form, " "))[1])
  out_var_ind <- as.numeric(which(colnames(data) == out_var))

  # split up data into training and validation sets
  train_data <- training(data)
  valid_data <- validation(data)

  # fit linear model on training set and predict on validation set
  mod <- lm(as.formula(reg_form), data = train_data)
  preds <- predict(mod, newdata = valid_data)

  # capture results to be returned as output
  out <- list(
    coef = data.frame(t(coef(mod))),
    SE = ((preds - valid_data[, out_var_ind])^2)
  )
  return(out)
}

plan(sequential)
time_seq <- system.time({
  results_seq <- cross_validate(
    cv_fun = cv_lm, folds = folds, data = mtcars,
    reg_form = "mpg ~ ."
  )
})

plan(multicore)
time_mc <- system.time({
  results_mc <- cross_validate(
    cv_fun = cv_lm, folds = folds, data = mtcars,
    reg_form = "mpg ~ ."
  )
})

if (availableCores() > 1) {
  time_mc["elapsed"] < 1.2 * time_seq["elapsed"]
}
}

=head1 NAME

Origami - command-line management tool for Folding @ Home clients.

=head1 SYNOPSIS

origami {install | archive | restore | erase | name | status | monitor | proxy | start | stop | version | help} [<options>...]

origami {deploy | network-archive | network-restore | network-name | network-monitor | network-name | network-start | network-stop} [<options>...]

origami help

origami version

=head1 DESCRIPTION

Origami allows you to install, monitor, archive or restore Folding @ Home data
on your local machine.  It also allows for central deployment, monitoring,
archiving, restoration and updating the username and team data via the network
to ssh-accessible machines. 

Origami supports cron-based scheduling, 32bit or 64bit and network proxy settings.
Install Stanfords Folding @ Home client to a single machine or network of machines.

=head1 COMMAND-LINE ACTIONS

The first argument which does not begin with a hyphen ("-") is considered to be
an action that the program should perform.  If an action is not specified
B<Origami> will display the help output.

The following actions are available to the local machine:

=over 8

=item B<install>

Install Folding @ Home to the local machine.  This action will create a
dedicated user, storage location for Folding @ Home data, startup
script and request a work unit.

=item B<deploy>

Install Folding @ Home to remote machines as listed in the ~/.origamirc file.
This action will create a dedicated user, storage location for Folding @ Home
data, startup script and request a work unit on the remote machine.  SSH
access as the root user is required for the deploy action.

=item B<archive>

Archive Folding @ Home data for the local Folding @ Home client.  This action 
will stop the Folding @ Home client, archive the current progress into a
<hostname>-<date>.tar.bz2 format and store the data in ~/.origami/.

=item B<network-archive>

Archive Folding @ Home data on the remote Folding @ Home client as listed in the
~/.origamirc.  This action will stop the Folding @ Home client, archive the 
current progress into a <hostname>-<date>.tar.bz2 format and store the data in ~/.origami/.
	
=item B<restore>

Restore Folding @ Home data on the local machine.  This action will
require previously archived data (as generated by the archive action).

=item B<network-restore>

Restore Folding @ Home data on the remote Folding @ Home clients as listed in the
~/.origamirc.  This action will require previously archived data (as generated by
the network-archive action).

=item B<erase>

Erase current Folding @ Home installation.  Anything generated by the
install function will be removed and all Folding @ Home progress will
be lost.

=item B<name>

Update the username and team information for the local Folding @ Home
client.  It is generally preferred to define username and team at
installation with the install action.

=item B<network-name>

Update the username and team information for the remote Folding @ Home
clients.  It is generally preferred to define username and team at
installation with the deploy action.

=item B<status|monitor>

Display the current progress of the local Folding @ Home client.

=item B<network-status|network-monitor>

Disploy the current progress of the remote Folding @ Home clients as listed
in the ~/.origamirc.

=item B<proxy>

Update the proxy settings for the local Folding @ Home client.  This
action requires a port and hostname.  It is generally preferred to
define the proxy port and hostname at installation with the install
action.

=item B<start>

Start the local Folding @ Home client.  (This calls /etc/init.d/origami start)

=item B<network-start>

Start the remote Folding @ Home clients as listed in the ~/.origamirc.

=item B<stop>

Stop the local Folding @ Home client.  (This calls /etc/init.d/origami stop)
	
=item B<network-stop>

Stop the remote Folding @ Home clients as listed in the ~/.origamirc.

=item B<version>

Display the Origami version.

=item B<help>

Display the Origami help output.

=back

=head1 OPTIONS

=over

=item B<-u|--user=I<value>>

Define the USERNAME that Folding @ Home data should credit.  Default is Anonymous.

=item B<-t|--team=I<value>>

Define the TEAM NUMBER that Folding @ Home data should credit.  Default is 0.

=item B<-c|--cron=I<boolean>>

Activate cron scheduling for the Folding @ Home client.  Default is off (0).

=item B<-p|--proc=I<value>>

Define i386 or amd64 Folding @ Home client.  Default is i386.

=item B<-P|--port=I<value>>

Define the port to be used in conjunction with a proxy.  Default is 8080.

=item B<-H|--host=I<value>>

Define the IP or hostname to be used in conjunction with a proxy.  Default is localhost.

=back

=head1 EXAMPLES

=over

=item B<origami install>

This example will install the Folding @ Home client to the local machine, reporting data
as the default Anonymous user and default 0 Team.

=item B<origami install -u Joe -t 45104>

This example will install the Folding @ Home client to the local machine, reporting data
as the B<Joe> user as part of the 45104 (TeamUbuntu) Team.

=item B<origami deploy -u Joe -t 45104>

This example will install the Folding @ Home client to all machines listed in the ~/.origamirc
file, reporting data as the B<Joe> user as part of the 45104 (TeamUbuntu) Team.

=item B<origami install -u Joe -t 45104 -p amd64 -c1>

This example will install the Folding @ home client to the local machine, reporting data
as the B<Joe> user as part of the 45104 (TeamUbuntu) Team, using the amd64 Folding @ Home client
and toggling the cron option to auto-stop between the hours of 8:00am and 5:00pm.

=item B<origami network-status>

This example will display the current status of the Folding @ Home clients on all machines
listed in the ~/.origamirc file.

=back

=head1 BUGS

To find a list of current bugs, or to report a bug new bug please see https://bugs.launchpad.net/origami.

=head1 AUTHORS

B<Origami> was written by Christer Edwards <christer.edwards@ubuntu.com>. 
This manual page was written by Christer Edwards <christer.edwards@ubuntu.com>.

Both are released under the GNU General Public License, version 2 or later.

=cut

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{folds2foldvec}
\alias{folds2foldvec}
\title{Build a Fold Vector from a Fold Object}
\usage{
folds2foldvec(folds)
}
\arguments{
\item{folds}{A \code{fold} object as produced by \code{\link{make_folds}},
from which a \code{numeric} vector of the validation set fold IDs are
returned.}
}
\description{
For V-fold type cross-validation. This takes a fold object and returns a fold
vector (containing the validation set IDs) for use with other tools like
\code{\link[glmnet]{cv.glmnet}}.
}
\seealso{
Other fold generation functions: 
\code{\link{fold_from_foldvec}()},
\code{\link{fold_funs}},
\code{\link{make_folds}()},
\code{\link{make_repeated_folds}()}
}
\concept{fold generation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/combine_results.R
\name{combine_results}
\alias{combine_results}
\title{Combine Results from Different Folds}
\usage{
combine_results(results, combiners = NULL, smart_combiners = TRUE)
}
\arguments{
\item{results}{A \code{list} of \code{list}s, corresponding to each result,
with the inner \code{list}s corresponding to results from each fold.}

\item{combiners}{A \code{list} with the same names as results, containing
combiner function names or functions for each result.}

\item{smart_combiners}{A \code{logical} indicating whether combiners should
be guessed from the data type of the results if they are missing.}
}
\value{
A list of combined results.
}
\description{
Applies \code{\link{combiners}}: functions that collapse across a list of
similarly structured results, to a list of such lists.
}
\details{
In theory you should never call this function directly, because it
 is called automatically by cross_validate. The defaults, combiners guessed
 based on data type, should work in most cases.
}
\seealso{
\code{\link{combiners}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fold_funs.R
\name{check_id_and_time}
\alias{check_id_and_time}
\title{Check ID and Time Compatibility}
\usage{
check_id_and_time(id, time)
}
\arguments{
\item{id}{An optional vector of unique identifiers corresponding to the time
vector. These can be used to subset the time vector.}

\item{time}{An optional vector of integers of time points observed for each
subject in the sample.}
}
\description{
Check ID and Time Compatibility
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/combine_results.R
\name{guess_combiner}
\alias{guess_combiner}
\title{Flexible Guessing and Mapping for Combining Data Types}
\usage{
guess_combiner(result)
}
\arguments{
\item{result}{A single result; flexibly accepts several object classes.}
}
\value{
A function to combine a list of such results.
}
\description{
Maps data types into standard combiners that should be sensible.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_folds.R
\name{make_folds}
\alias{make_folds}
\title{Make List of Folds for cross-validation}
\usage{
make_folds(
  n = NULL,
  fold_fun = folds_vfold,
  cluster_ids = NULL,
  strata_ids = NULL,
  ...
)
}
\arguments{
\item{n}{- either an integer indicating the number of observations to
cross-validate over, or an object from which to guess the number of
observations; can also be computed from \code{strata_ids} or
\code{cluster_ids}.}

\item{fold_fun}{- A function indicating the cross-validation scheme to use.
See \code{\link{fold_funs}} for a list of possibilities.}

\item{cluster_ids}{- a vector of cluster ids. Clusters are treated as a unit
-- that is, all observations within a cluster are placed in either the
training or validation set.}

\item{strata_ids}{- a vector of strata ids. Strata are balanced: insofar as
possible the distribution in the sample should be the same as the
distribution in the training and validation sets.}

\item{...}{other arguments to be passed to \code{fold_fun}.}
}
\value{
A list of folds objects. Each fold consists of a list with a
 \code{training} index vector, a \code{validation} index vector, and a
 \code{fold_index} (its order in the list of folds).
}
\description{
Generates a list of folds for a variety of cross-validation schemes.
}
\seealso{
Other fold generation functions: 
\code{\link{fold_from_foldvec}()},
\code{\link{fold_funs}},
\code{\link{folds2foldvec}()},
\code{\link{make_repeated_folds}()}
}
\concept{fold generation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fold_funs.R
\name{fold_funs}
\alias{fold_funs}
\alias{folds_vfold}
\alias{folds_resubstitution}
\alias{folds_loo}
\alias{folds_montecarlo}
\alias{folds_bootstrap}
\alias{folds_rolling_origin}
\alias{folds_rolling_window}
\alias{folds_rolling_origin_pooled}
\alias{folds_rolling_window_pooled}
\alias{folds_vfold_rolling_origin_pooled}
\alias{folds_vfold_rolling_window_pooled}
\title{Cross-Validation Schemes}
\usage{
folds_vfold(n, V = 10L)

folds_resubstitution(n)

folds_loo(n)

folds_montecarlo(n, V = 1000L, pvalidation = 0.2)

folds_bootstrap(n, V = 1000L)

folds_rolling_origin(n, first_window, validation_size, gap = 0L, batch = 1L)

folds_rolling_window(n, window_size, validation_size, gap = 0L, batch = 1L)

folds_rolling_origin_pooled(
  n,
  t,
  id = NULL,
  time = NULL,
  first_window,
  validation_size,
  gap = 0L,
  batch = 1L
)

folds_rolling_window_pooled(
  n,
  t,
  id = NULL,
  time = NULL,
  window_size,
  validation_size,
  gap = 0L,
  batch = 1L
)

folds_vfold_rolling_origin_pooled(
  n,
  t,
  id = NULL,
  time = NULL,
  V = 10L,
  first_window,
  validation_size,
  gap = 0L,
  batch = 1L
)

folds_vfold_rolling_window_pooled(
  n,
  t,
  id = NULL,
  time = NULL,
  V = 10L,
  window_size,
  validation_size,
  gap = 0L,
  batch = 1L
)
}
\arguments{
\item{n}{An integer indicating the number of observations.}

\item{V}{An integer indicating the number of folds.}

\item{pvalidation}{A \code{numeric} indicating the proportion of observation
to be placed in the validation fold.}

\item{first_window}{An integer indicating the number of observations in the
first training sample.}

\item{validation_size}{An integer indicating the number of points in the
validation samples; should be equal to the largest forecast horizon.}

\item{gap}{An integer indicating the number of points not included in the
training or validation samples. The default is zero.}

\item{batch}{An integer indicating increases in the number of time points
added to the training set in each iteration of cross-validation. Applicable
for larger time-series. The default is one.}

\item{window_size}{An integer indicating the number of observations in each
training sample.}

\item{t}{An integer indicating the total amount of time to consider per
time-series sample.}

\item{id}{An optional vector of unique identifiers corresponding to the time
vector. These can be used to subset the time vector.}

\item{time}{An optional vector of integers of time points observed for each
subject in the sample.}
}
\value{
A list of \code{Fold}s.
}
\description{
These functions represent different cross-validation schemes that can be
used with \pkg{origami}. They should be used as options for the
\code{fold_fun} argument to \code{\link{make_folds}}, which will call the
requested function specify \code{n}, based on its arguments, and pass any
remaining arguments (e.g. \code{V} or \code{pvalidation}) on.
}
\seealso{
Other fold generation functions: 
\code{\link{fold_from_foldvec}()},
\code{\link{folds2foldvec}()},
\code{\link{make_folds}()},
\code{\link{make_repeated_folds}()}
}
\concept{fold generation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{wrap_in_try}
\alias{wrap_in_try}
\title{Wrap a Function in a Try Statement}
\usage{
wrap_in_try(fun, ...)
}
\arguments{
\item{fun}{A \code{function} to be wrapped in a \code{try} statement.}

\item{...}{Additional arguments passed to the previous argument \code{fun}.}
}
\description{
Function factory that generates versions of functions wrapped in \code{try}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{fold_from_foldvec}
\alias{fold_from_foldvec}
\title{Build a Fold Object from a Fold Vector}
\usage{
fold_from_foldvec(v, folds)
}
\arguments{
\item{v}{An identifier of the fold in which observations fall for
cross-validation.}

\item{folds}{A vector of the fold status for each observation for
cross-validation.}
}
\description{
For V-fold type cross-validation. This takes a fold vector (validation set
IDs) and builds a fold object for fold V.
}
\seealso{
Other fold generation functions: 
\code{\link{fold_funs}},
\code{\link{folds2foldvec}()},
\code{\link{make_folds}()},
\code{\link{make_repeated_folds}()}
}
\concept{fold generation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fold.R
\name{fold_helpers}
\alias{fold_helpers}
\alias{training}
\alias{validation}
\alias{fold_index}
\title{Fold Helpers}
\usage{
training(x = NULL, fold = NULL)

validation(x = NULL, fold = NULL)

fold_index(x = NULL, fold = NULL)
}
\arguments{
\item{x}{an object to be indexed by a training set, validation set, or fold
index. If missing, the index itself will be returned.}

\item{fold}{Fold; the fold used to do the indexing. If missing, \code{fold}
will be pulled from the calling environment, if available.}
}
\value{
The elements of \code{x} corresponding to the indexes, or the
 indexes themselves if \code{x} is missing.
}
\description{
Accessors and indexers for the different parts of a fold.
}
\seealso{
\code{\link{make_fold}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_folds.R
\name{id_folds_to_folds}
\alias{id_folds_to_folds}
\title{Convert ID Folds to Observation Folds}
\usage{
id_folds_to_folds(idfolds, cluster_ids)
}
\arguments{
\item{idfolds}{folds that subset ids}

\item{cluster_ids}{a vector of cluster ids indicating which observations are
in which clusters}
}
\description{
This function convertsf olds that subset ids to folds that subset
 observations
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_folds.R
\name{make_repeated_folds}
\alias{make_repeated_folds}
\title{Repeated Cross-Validation}
\usage{
make_repeated_folds(repeats, ...)
}
\arguments{
\item{repeats}{An integer indicating the number of repeats.}

\item{...}{Arguments passed to \code{\link{make_folds}}.}
}
\description{
Implementation of repeated window cross-validation: generates fold objects
for repeated cross-validation by making repeated calls to
\code{\link{make_folds}} and concatenating the results.
}
\seealso{
Other fold generation functions: 
\code{\link{fold_from_foldvec}()},
\code{\link{fold_funs}},
\code{\link{folds2foldvec}()},
\code{\link{make_folds}()}
}
\concept{fold generation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fold.R
\name{make_fold}
\alias{make_fold}
\title{Fold}
\usage{
make_fold(v, training_set, validation_set)
}
\arguments{
\item{v}{An integer index of folds in the larger scheme.}

\item{training_set}{An integer vector of indexes corresponding to the
training set.}

\item{validation_set}{An integer vector of indexes corresponding to the
validation set.}
}
\value{
A list containing these elements.
}
\description{
Functions to make a fold. Current representation is a simple \code{list}.
}
\seealso{
\code{\link{fold_helpers}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/combine_results.R
\name{combiners}
\alias{combiners}
\alias{combiner_rbind}
\alias{combiner_c}
\alias{combiner_factor}
\alias{combiner_array}
\title{Combiners}
\usage{
combiner_rbind(x)

combiner_c(x)

combiner_factor(x)

combiner_array(x)
}
\arguments{
\item{x}{A \code{list} of similar results to be combined.}
}
\value{
A combined results object.
}
\description{
Combiners are functions that collapse across a list of similarly structured
results. These are standard idioms for combining lists of certain data types.
}
